-- BeneCast by Skurel
-- v 2.1.1
-- Code heavily referenced from Danboo's Healomatic/CastParty. Thank you Danboo!
-- Makes clickable buttons that casts beneficial spells on party members and the current target

-- *****************************************************************************
-- BeneCast Constants
-- *****************************************************************************
-- Maximum number of buttons in a BeneCast Panel
local BENECAST_MAX_NUMBER_OF_BUTTONS = 10;
-- This is used to reset the configuration if severely outdated
local BENECAST_VERSION = 2.0;
-- Constants used for Blizzard Raid UI pullout frames
local BLIZZ_MAX_RAID_PULLOUT_BUTTONS = 15;
local BLIZZ_MAX_RAID_PULLOUT_FRAMES = 12;

-- *****************************************************************************
-- BeneCast Raid Tables
-- *****************************************************************************
local BENECAST_RAID_LIST = {};
local BENECAST_RAID_SUBGROUPS = { 0, 0, 0, 0, 0, 0, 0, 0 };

-- *****************************************************************************
-- BeneCast Sting Constants - Load from the appropriate Localization table
-- *****************************************************************************
local BENECAST_STRINGS = BeneCast_Localization[GetLocale()];

-- *****************************************************************************
-- BeneCast UI Tables
-- *****************************************************************************

-- Table to hold all the OptionCheckButton information
-- table value: Text of the checkbutton label
-- index: ID of the checkbutton
-- cvar: Option modified in BeneCastConfig
-- tooltipText: Text for the tooltip of the checkbutton
local BeneCastOptionFrameCheckButtons = {};
BeneCastOptionFrameCheckButtons['TEXT_DAMAGE_BASED_HEALS'] = { index = 1, cvar = 'DmgBasedHeal', tooltipText = 'TEXT_DAMAGE_BASED_HEALS_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_OVERHEAL'] = { index = 2, cvar = 'Overheal', tooltipText = 'TEXT_OVERHEAL_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_SHOWALLBUFFS'] = { index = 3, cvar = 'ShowAllBuffs', tooltipText = 'TEXT_SHOWALLBUFFS_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_HIDE_MINIMAP_BUTTON'] = { index = 4, cvar = 'HideMinimap', tooltipText = 'TEXT_HIDE_MINIMAP_BUTTON_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_TOOLTIPS'] = { index = 5, cvar = 'ShowTooltips', tooltipText = 'TEXT_TOOLTIPS_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_TOOLTIPS_NAME'] = { index = 6, cvar = 'ShowTooltipsName', tooltipText = 'TEXT_TOOLTIPS_NAME_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_UNLOCK_BUTTONS'] = { index = 7, cvar = 'Unlock', tooltipText = 'TEXT_UNLOCK_BUTTONS_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_NOTIFICATION'] = { index = 8, cvar = 'Notification', tooltipText = 'TEXT_PARTY_NOTIFICATION_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_SELF_CHANNEL'] = { index = 9, cvar = 'SelfNotify', tooltipText = 'TEXT_SELF_CHANNEL_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_PARTY_CHANNEL'] = { index = 10, cvar = 'PartyNotify', tooltipText = 'TEXT_PARTY_CHANNEL_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_RAID_CHANNEL'] = { index = 11, cvar = 'RaidNotify', tooltipText = 'TEXT_RAID_CHANNEL_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_WHISPER_TARGET'] = { index = 12, cvar = 'TargetNotify', tooltipText = 'TEXT_WHISPER_TARGET_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_NOTIFY_SELF_CASTS'] = { index = 13, cvar = 'SelfCasts', tooltipText = 'TEXT_NOTIFY_SELF_CASTS_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_NOTIFY_RANK'] = { index = 14, cvar = 'NotifyRank', tooltipText = 'TEXT_NOTIFY_RANK_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_NOTIFY_TIME'] = { index = 15, cvar = 'NotifyTime', tooltipText = 'TEXT_NOTIFY_TIME_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_NOTIFY_HEALS'] = { index = 16, cvar = 'NotifyHeal', tooltipText = 'TEXT_NOTIFY_HEALS_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_NOTIFY_MAX_HEALS'] = { index = 17, cvar = 'NotifyMaxHeal', tooltipText = 'TEXT_NOTIFY_MAX_HEALS_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_NOTIFY_CURES'] = { index = 18, cvar = 'NotifyCure', tooltipText = 'TEXT_NOTIFY_CURES_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_NOTIFY_BUFFS'] = { index = 19, cvar = 'NotifyBuff', tooltipText = 'TEXT_NOTIFY_BUFFS_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_NOTIFY_USER_CHANNEL'] = { index = 20, cvar = 'UserNotify', tooltipText = 'TEXT_NOTIFY_USER_CHANNEL_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_NOTIFY_RES'] = { index = 21, cvar = 'NotifyRes', tooltipText = 'TEXT_NOTIFY_RES_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_PLAYER_AS_DEFAULT'] = { index = 22, cvar = 'PlayerAsDefault', tooltipText = 'TEXT_PLAYER_AS_DEFAULT_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_SHOW_PETS'] = { index = 23, cvar = 'ShowPets', tooltipText = 'TEXT_SHOW_PETS_TOOLTIP' };
BeneCastOptionFrameCheckButtons['TEXT_SHOW_RAID'] = { index = 24, cvar = 'ShowRaid', tooltipText = 'TEXT_SHOW_RAID_TOOLTIP' };

-- Table to hold all the OptionCheckButton information
-- text: Text of the slider label
-- index: ID of the slider
-- cvar: Option modified in BeneCastConfig
-- minValue: Minimum value of the slider
-- maxValue: Maximum value of the slider
-- valueStep: Step value of the slider when being changed
-- tooltipText: Text for the tooltip of the checkbutton
local BeneCastOptionFrameSliders = {
	{ index = 1, text = 'TEXT_BUTTON_SIZE', cvar = 'ButtonSize', minValue = 8, maxValue = 36, valueStep = 1, tooltipText = 'TEXT_BUTTONSIZESLIDER_TOOLTIP' },
	{ index = 2, text = 'TEXT_RANKS_TO_OVERHEAL', cvar = 'OverhealSlider', minValue = 0, maxValue = 10, valueStep = 1, tooltipText = 'TEXT_OVERHEALSLIDER_TOOLTIP' },
	{ index = 3, text = 'TEXT_BUTTON_NUMBER', cvar = 'ButtonNumber', minValue = 1, maxValue = 10, valueStep = 1, tooltipText = 'TEXT_BUTTONNUMBERSLIDER_TOOLTIP' },
};

-- Table to hold the different subframe names in the BeneCastOptions frame
local BeneCastOptionFrameSubframes = {'BeneCastPlayerFrame', 'BeneCastClass1Frame', 'BeneCastClass2Frame', 'BeneCastClass3Frame', 'BeneCastClass4Frame', 'BeneCastClass5Frame', 'BeneCastClass6Frame', 'BeneCastClass7Frame', 'BeneCastClass8Frame', 'BeneCastSetupFrame', 'BeneCastNotificationFrame', 'BeneCastRaidFrame', };

-- *****************************************************************************
-- Global variables
-- *****************************************************************************

-- Array to hold possible targets
local _targets = {'player', 'party1', 'party2', 'party3', 'party4', 'raid1', 'raid2', 'raid3', 'raid4', 'raid5', 'raid6', 'raid7', 'raid8', 'raid9', 'raid10', 'raid11', 'raid12', 'raid13', 'raid14', 'raid15', 'raid16', 'raid17', 'raid18', 'raid19', 'raid20', 'raid21', 'raid22', 'raid23', 'raid24', 'raid25', 'raid26', 'raid27', 'raid28', 'raid29', 'raid30', 'raid31', 'raid32', 'raid33', 'raid34', 'raid35', 'raid36', 'raid37', 'raid38', 'raid39', 'raid40', 'pet', 'partypet1', 'partypet2', 'partypet3', 'partypet4', 'raidpet1', 'raidpet2', 'raidpet3', 'raidpet4', 'raidpet5', 'raidpet6', 'raidpet7', 'raidpet8', 'raidpet9', 'raidpet10', 'raidpet11', 'raidpet12', 'raidpet13', 'raidpet14', 'raidpet15', 'raidpet16', 'raidpet17', 'raidpet18', 'raidpet19', 'raidpet20', 'raidpet21', 'raidpet22', 'raidpet23', 'raidpet24', 'raidpet25', 'raidpet26', 'raidpet27', 'raidpet28', 'raidpet29', 'raidpet30', 'raidpet31', 'raidpet32', 'raidpet33', 'raidpet34', 'raidpet35', 'raidpet36', 'raidpet37', 'raidpet38', 'raidpet39', 'raidpet40', 'target',};

-- Array to assign id numbers based on class
local _classes = {
	[BENECAST_STRINGS.CLASS_DRUID]		= 2,
	[BENECAST_STRINGS.CLASS_MAGE]		= 4,
	[BENECAST_STRINGS.CLASS_HUNTER]		= 3,
	[BENECAST_STRINGS.CLASS_PRIEST]		= 5,
	[BENECAST_STRINGS.CLASS_ROGUE]		= 6,
	[BENECAST_STRINGS.CLASS_WARLOCK]	= 7,
	[BENECAST_STRINGS.CLASS_WARRIOR]	= 8,
	[BENECAST_STRINGS.CLASS_SHAMAN]		= 9,
	[BENECAST_STRINGS.CLASS_PALADIN]	= 9,
}

-- Table used to sort spelltypes
local _spellsort = { 'res', 'efficient', 'emergency', 'instant', 'loh', 'group', 'poison', 'disease', 'magic', 'curse', 'buff1', 'buff2', 'buff3', 'buff4', 'buff5', 'buff6', 'groupbuff', 'buffparty1', 'buffparty2', 'buffparty3', 'buffparty4', 'buffparty5', 'buffparty6', 'buffparty7', 'buffparty8', 'buffparty9', 'selfbuff1', 'selfbuff2', 'selfbuff3', 'selfbuff4', 'selfbuff5', 'selfbuff6', 'selfbuff7', 'selfbuff8', 'selfbuff9', 'selfbuff10', 'selfbuff11', 'selfbuff12', 'selfbuff13', 'weaponenchant1', 'weaponenchant2', 'weaponenchant3', 'weaponenchant4', };

-- Table used to determine the buffs in effect in UpdateButtons
local _buffs_in_effect = {};
_buffs_in_effect[1] = { ineffect = nil, fade = nil, name = 'buff1' };
_buffs_in_effect[2] = { ineffect = nil, fade = nil, name = 'buff2' };
_buffs_in_effect[3] = { ineffect = nil, fade = nil, name = 'buff3' };
_buffs_in_effect[4] = { ineffect = nil, fade = nil, name = 'buff4' };
_buffs_in_effect[5] = { ineffect = nil, fade = nil, name = 'buff5' };
_buffs_in_effect[6] = { ineffect = nil, fade = nil, name = 'buff6' };
_buffs_in_effect[7] = { ineffect = nil, fade = nil, name = 'groupbuff' };
_buffs_in_effect[8] = { ineffect = nil, fade = nil, name = 'buffparty1' };
_buffs_in_effect[9] = { ineffect = nil, fade = nil, name = 'buffparty2' };
_buffs_in_effect[10] = { ineffect = nil, fade = nil, name = 'buffparty3' };
_buffs_in_effect[11] = { ineffect = nil, fade = nil, name = 'buffparty4' };
_buffs_in_effect[12] = { ineffect = nil, fade = nil, name = 'buffparty5' };
_buffs_in_effect[13] = { ineffect = nil, fade = nil, name = 'buffparty6' };
_buffs_in_effect[14] = { ineffect = nil, fade = nil, name = 'buffparty7' };
_buffs_in_effect[15] = { ineffect = nil, fade = nil, name = 'buffparty8' };
_buffs_in_effect[16] = { ineffect = nil, fade = nil, name = 'buffparty9' };
_buffs_in_effect[17] = { ineffect = nil, fade = nil, name = 'selfbuff1' };
_buffs_in_effect[18] = { ineffect = nil, fade = nil, name = 'selfbuff2' };
_buffs_in_effect[19] = { ineffect = nil, fade = nil, name = 'selfbuff3' };
_buffs_in_effect[20] = { ineffect = nil, fade = nil, name = 'selfbuff4' };
_buffs_in_effect[21] = { ineffect = nil, fade = nil, name = 'selfbuff5' };
_buffs_in_effect[22] = { ineffect = nil, fade = nil, name = 'selfbuff6' };
_buffs_in_effect[23] = { ineffect = nil, fade = nil, name = 'selfbuff7' };
_buffs_in_effect[24] = { ineffect = nil, fade = nil, name = 'selfbuff8' };
_buffs_in_effect[25] = { ineffect = nil, fade = nil, name = 'selfbuff9' };
_buffs_in_effect[26] = { ineffect = nil, fade = nil, name = 'selfbuff10' };
_buffs_in_effect[27] = { ineffect = nil, fade = nil, name = 'selfbuff11' };
_buffs_in_effect[28] = { ineffect = nil, fade = nil, name = 'selfbuff12' };
_buffs_in_effect[29] = { ineffect = nil, fade = nil, name = 'selfbuff13' };

-- Healing threshold from max
local _threshold = .9;

-- Class of the player
local _class;

-- Name of the player
local _name;

-- Party notification stuff
local _spellbeingcast = nil;
local _spelltarget = nil;
local _spellismax = nil;
local _spellstarted = nil;
local _spellstopped = nil;
local _sentmessage = nil;

-- Weapon Enchant status
local _weaponenchant = nil;

-- Clearcasting status
local _clearcasting = nil;

-- Tables for interpolating health of targets outside the group
local _class_health = {
	[BENECAST_STRINGS.CLASS_DRUID]	= { ['1'] = 30, ['60'] = 3000 },
	[BENECAST_STRINGS.CLASS_MAGE]	= { ['1'] = 30, ['60'] = 3000 },
	[BENECAST_STRINGS.CLASS_HUNTER]	= { ['1'] = 30, ['60'] = 4000 },
	[BENECAST_STRINGS.CLASS_PALADIN]= { ['1'] = 30, ['60'] = 4500 },
	[BENECAST_STRINGS.CLASS_PRIEST]	= { ['1'] = 30, ['60'] = 3000 },
	[BENECAST_STRINGS.CLASS_ROGUE]	= { ['1'] = 30, ['60'] = 3500 },
	[BENECAST_STRINGS.CLASS_SHAMAN]	= { ['1'] = 30, ['60'] = 4000 },
	[BENECAST_STRINGS.CLASS_WARLOCK]= { ['1'] = 30, ['60'] = 4000 },
	[BENECAST_STRINGS.CLASS_WARRIOR]= { ['1'] = 30, ['60'] = 5000 },
};

-- Spell Level for spells, used for determining what can be cast on the target
-- Spell Level is the level at which you are able to train the spell
-- The target level must be >= spell level - 10
local _spell_level = {
	[BENECAST_STRINGS.CLASS_DRUID] = {
		['assist'] = { 1 },
		['res'] = { 1, 1, 1, 1, 1, },
		['efficient'] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
		['emergency'] = { 12, 18, 24, 30, 36, 42, 48, 54, 60 },
		['instant'] = { 4, 10, 16, 22, 28, 34, 40, 46, 52, 58 },
		['group'] = { 1, 1, 1, 1 },
		['poison'] = { 1, 1 },
		['curse'] = { 1 },
		['selfbuff1'] = { 1 },
		['selfbuff2'] = { 1, 18, 28, 38, 48, 58 },
		['selfbuff3'] = { 44 },
		['selfbuff4'] = { 1 },
		['buff1'] = { 1, 10, 20, 30, 40, 50, 60 },
		['buff2'] = { 6, 14, 24, 34, 44, 54 },
		['buff3'] = { 40 },
		['groupbuff'] = { 50, 60 },
	},
	[BENECAST_STRINGS.CLASS_HUNTER] = {
		['assist'] = { 1 },
		['selfbuff1'] = { 1 },
		['selfbuff2'] = { 1 },
		['selfbuff3'] = { 1, 1, 1, 1, 1, 1, 1 },
		['selfbuff4'] = { 1 },
		['selfbuff5'] = { 1 },
		['selfbuff6'] = { 1, 1 },
		['selfbuff7'] = { 1, 1, 1 },
		['selfbuff8'] = { 1 },
		['selfbuff9'] = { 1 },
		['selfbuff10'] = { 1 },
	},
	[BENECAST_STRINGS.CLASS_MAGE] = {
		['assist'] = { 1 },
		['curse'] = { 1, 1 },
		['selfbuff1'] = { 1, 10, 20, 30, 40, 50, 60 },
		['selfbuff2'] = { 34, 46, 58 },
		['selfbuff3'] = { 1 },
		['selfbuff4'] = { 1 },
		['selfbuff5'] = { 1 },
		['selfbuff6'] = { 1 },
		['selfbuff7'] = { 1 },
		['selfbuff8'] = { 1, 46, 52, 58 },
		['selfbuff9'] = { 1 },
		['selfbuff10'] = { 12 },
		['selfbuff11'] = { 20, 28, 36, 44, 52, 60 },
		['selfbuff12'] = { 20, 30, 40, 50, 60 },
		['selfbuff13'] = { 22, 32, 42, 52 },
		['buffparty1'] = { 18, 30, 42, 54 },
		['buffparty2'] = { 12, 24, 36, 48, 60 },
		['buff1'] = { 1, 14, 28, 42, 56 },
		['groupbuff'] = { 56 },
	},
	[BENECAST_STRINGS.CLASS_PALADIN] = {
		['assist'] = { 1 },
		['res'] = { 1, 1, 1, 1, 1, },
		['efficient'] = { 1, 1, 1, 1, 1, 1, 1, 1 },
		['emergency'] = { 1, 1, 1, 1, 1, 1 },
		['instant'] = { 1, 1, 1 },
		['loh'] = { 1, 1, 1 },
		['poison'] = { 1, 1 },
		['selfbuff1'] = { 6, 18, 34, 50 },
		['selfbuff2'] = { 16, 26, 36, 46, 56, },
		['selfbuff3'] = { 22 },
		['selfbuff4'] = { 30, 40, 50, 60 },
		['selfbuff5'] = { 1, 10, 18, 26, 34, 42, 50, 58 },
		['selfbuff6'] = { 38, 48, 58 },
		['selfbuff7'] = { 6, 12, 22, 32, 42, 52 },
		['selfbuff8'] = { 1, 30, 40, 50, 60 },
		['selfbuff9'] = { 1, 40, 50, 60 },
		['selfbuff10'] = { 1 },
		['selfbuff11'] = { 1 },
		['buff1'] = { 4, 12, 22, 32, 42, 52 },
		['buff2'] = { 18 },
		['buff3'] = { 40, 50, 60 },
		['buff4'] = { 1, 30, 40, 50, 60 },
		['buff5'] = { 14, 24, 34, 44, 54 },
		['buff6'] = { 40 },
		['buffparty1'] = { 10, 24, 38, 46, 54 },
		['buffparty2'] = { 1 },
		['buffparty3'] = { 46, 54 },
		['buffparty4'] = { 30 },
		['buffparty5'] = { 52, 60 },
		['buffparty6'] = { 60 },
		['buffparty7'] = { 60 },
		['buffparty8'] = { 54, 60 },
		['buffparty9'] = { 60 },
		['buffparty10'] = { 60, 60, 60 },
	},
	[BENECAST_STRINGS.CLASS_PRIEST] = {
		['assist'] = { 1 },
		['res'] = { 1, 1, 1, 1, 1, },
		['efficient'] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
		['emergency'] = { 1, 1, 1, 1, 1, 1, 1 },
		['instant'] = { 8, 14, 20, 26, 32, 38, 44, 50, 56 },
		['group'] = { 1, 1, 1, 1 },
		['disease'] = { 1, 1 },
		['magic'] = { 1, 1 },
		['selfbuff1'] = { 1 },
		['selfbuff2'] = { 1 },
		['selfbuff3'] = { 1, 1, 1, 1, 1, 1, 1 },
		['selfbuff4'] = { 1, 30, 40, 50, 60 },
		['selfbuff5'] = { 8, 20, 30, 40, 50, 60 },
		['selfbuff6'] = { 1 },
		['selfbuff7'] = { 1, 28, 36, 44, 52, 60 },
		['selfbuff8'] = { 12, 20, 30, 40, 50, 60 },
		['selfbuff9'] = { 10, 20, 30, 40, 50, 60 },
		['buffparty1'] = { 6, 12, 18, 24, 30, 36, 42, 48, 54, 60 },
		['buff1'] = { 30, 42, 54 },
		['buff2'] = { 1, 12, 24, 36, 48, 60 },
		['buff3'] = { 30, 42, 56 },
		['buff4'] = { 20 },
		['groupbuff'] = { 48, 60 },
		['weaponenchant1'] = { 20, 30, 40, 50, 60 },
	},
	[BENECAST_STRINGS.CLASS_ROGUE] = {
		['assist'] = { 1 },
		['selfbuff1'] = { 1 },
		['selfbuff2'] = { 1 },
		['selfbuff3'] = { 1 },
		['selfbuff4'] = { 1 },
		['selfbuff5'] = { 1 },
		['selfbuff6'] = { 1 },
		['selfbuff7'] = { 1 },
		['selfbuff8'] = { 1, 1 },
	},
	[BENECAST_STRINGS.CLASS_SHAMAN] = {
		['assist'] = { 1 },
		['res'] = { 1, 1, 1, 1, 1, },
		['efficient'] = { 1, 1, 1, 1, 1, 1, 1, 1, 1 },
		['emergency'] = { 1, 1, 1, 1, 1, 1 },
		['group'] = { 1, 1, 1 },
		['poison'] = { 1 },
		['disease'] = { 1 },
		['selfbuff1'] = { 1 },
		['selfbuff2'] = { 8, 16, 24, 32, 40, 48, 56 },
		['buff1'] = { 22 },
		['buff2'] = { 28 },
		['weaponenchant1'] = { 1, 8, 16, 24, 34, 44, 54 },
		['weaponenchant2'] = { 10, 18, 26, 36, 46, 56 },
		['weaponenchant3'] = { 20, 28, 38, 48, 58 },
		['weaponenchant4'] = { 30, 40, 50, 60 },
	},
	[BENECAST_STRINGS.CLASS_WARLOCK] = {
		['assist'] = { 1 },
		['selfbuff1'] = { 1, 10, 20, 30, 40, 50, 60 },
		['selfbuff2'] = { 1, 1, 1 },
		['selfbuff3'] = { 1 },
		['selfbuff4'] = { 1 },
		['selfbuff5'] = { 1 },
		['selfbuff6'] = { 32, 42, 52 },
		['buff1'] = { 26, 38, 50 },
		['buff2'] = { 16 },
	},
	[BENECAST_STRINGS.CLASS_WARRIOR] = {
		['assist'] = { 1 },
		['selfbuff1'] = { 1 },
		['selfbuff2'] = { 1 },
		['selfbuff3'] = { 1 },
		['selfbuff4'] = { 1 },
		['selfbuff5'] = { 1 },
		['selfbuff6'] = { 1, 1, 1, 1, 1, 1 },
		['selfbuff7'] = { 1 },
		['selfbuff8'] = { 1,},
		['selfbuff9'] = { 1 },
		['selfbuff10'] = { 1 },
	},
};

-- Storage for spell data
local _spell_data = {};

-- Maping of spell lines to casting categories
local _spell_types = {
	[BENECAST_STRINGS.CLASS_DRUID] = {
		[BENECAST_STRINGS.RES_REBIRTH] = 'res',
		[BENECAST_STRINGS.HEAL_HEALING_TOUCH] = 'efficient',
		[BENECAST_STRINGS.HEAL_REGROWTH] = 'emergency',
		[BENECAST_STRINGS.HEAL_REJUVENATION] = 'instant',
		[BENECAST_STRINGS.HEAL_TRANQUILITY] = 'group',
		[BENECAST_STRINGS.CURE_POISON_1] = 'poison',
		[BENECAST_STRINGS.CURE_POISON_2] = 'poison',
		[BENECAST_STRINGS.CURE_CURSE_2] = 'curse',
		[BENECAST_STRINGS.SELFBUFF_NATURES_SWIFTNESS] = 'selfbuff1',
		[BENECAST_STRINGS.SELFBUFF_NATURES_GRASP] = 'selfbuff2',
		[BENECAST_STRINGS.SELFBUFF_BARKSKIN] = 'selfbuff3',
		[BENECAST_STRINGS.SELFBUFF_OMEN_OF_CLARITY] = 'selfbuff4',
		[BENECAST_STRINGS.BUFF_MARK_OF_THE_WILD] = 'buff1',
		[BENECAST_STRINGS.BUFF_THORNS] = 'buff2',
		[BENECAST_STRINGS.BUFF_INNERVATE] = 'buff3',
		[BENECAST_STRINGS.GROUPBUFF_GIFT_OF_THE_WILD] = 'groupbuff',
	},
	[BENECAST_STRINGS.CLASS_HUNTER] = {
		[BENECAST_STRINGS.SELFBUFF_ASPECT_OF_THE_BEAST] = 'selfbuff1',
		[BENECAST_STRINGS.SELFBUFF_ASPECT_OF_THE_CHEETAH] = 'selfbuff2',
		[BENECAST_STRINGS.SELFBUFF_ASPECT_OF_THE_HAWK] = 'selfbuff3',
		[BENECAST_STRINGS.SELFBUFF_ASPECT_OF_THE_MONKEY] = 'selfbuff4',
		[BENECAST_STRINGS.SELFBUFF_ASPECT_OF_THE_PACK] = 'selfbuff5',
		[BENECAST_STRINGS.SELFBUFF_ASPECT_OF_THE_WILD] = 'selfbuff6',
		[BENECAST_STRINGS.SELFBUFF_TRUESHOT_AURA] = 'selfbuff7',
		[BENECAST_STRINGS.SELFBUFF_DETERRENCE] = 'selfbuff8',
		[BENECAST_STRINGS.SELFBUFF_RAPID_FIRE] = 'selfbuff9',
		[BENECAST_STRINGS.SELFBUFF_FEIGN_DEATH] = 'selfbuff10',
	},
	[BENECAST_STRINGS.CLASS_MAGE] = {
		[BENECAST_STRINGS.CURE_CURSE_1] = 'curse',
		[BENECAST_STRINGS.SELFBUFF_FROST_ARMOR] = 'selfbuff1',
		[BENECAST_STRINGS.SELFBUFF_ICE_ARMOR] = 'selfbuff1',
		[BENECAST_STRINGS.SELFBUFF_MAGE_ARMOR] = 'selfbuff2',
		[BENECAST_STRINGS.SELFBUFF_EVOCATION] = 'selfbuff3',
		[BENECAST_STRINGS.SELFBUFF_PRESENCE_OF_MIND] = 'selfbuff4',
		[BENECAST_STRINGS.SELFBUFF_ARCANE_POWER] = 'selfbuff5',
		[BENECAST_STRINGS.SELFBUFF_COMBUSTION] = 'selfbuff6',
		[BENECAST_STRINGS.SELFBUFF_ICE_BLOCK] = 'selfbuff7',
		[BENECAST_STRINGS.SELFBUFF_ICE_BARRIER] = 'selfbuff8',
		[BENECAST_STRINGS.SELFBUFF_COLD_SNAP] = 'selfbuff9',
		[BENECAST_STRINGS.SELFBUFF_SLOW_FALL] = 'selfbuff10',
		[BENECAST_STRINGS.SELFBUFF_MANA_SHIELD] = 'selfbuff11',
		[BENECAST_STRINGS.SELFBUFF_FIRE_WARD] = 'selfbuff12',
		[BENECAST_STRINGS.SELFBUFF_FROST_WARD] = 'selfbuff13',
		[BENECAST_STRINGS.PARTYBUFF_AMPLIFY_MAGIC] = 'buffparty1',
		[BENECAST_STRINGS.PARTYBUFF_DAMPEN_MAGIC] = 'buffparty2',
		[BENECAST_STRINGS.BUFF_ARCANE_INTELLECT] = 'buff1',
		[BENECAST_STRINGS.GROUPBUFF_ARCANE_BRILLIANCE] = 'groupbuff',
	},
	[BENECAST_STRINGS.CLASS_PALADIN] = {
		[BENECAST_STRINGS.RES_REDEMPTION] = 'res',
		[BENECAST_STRINGS.HEAL_HOLY_LIGHT] = 'efficient',
		[BENECAST_STRINGS.HEAL_FLASH_OF_LIGHT] = 'emergency',
		[BENECAST_STRINGS.HEAL_HOLY_SHOCK] = 'instant',
		[BENECAST_STRINGS.HEAL_LAY_ON_HANDS] = 'loh',
		[BENECAST_STRINGS.CURE_PURIFY_1] = 'poison',
		[BENECAST_STRINGS.CURE_CLEANSE_1] = 'poison',
		[BENECAST_STRINGS.SELFBUFF_DIVINE_PROTECTION] = 'selfbuff1',
		[BENECAST_STRINGS.SELFBUFF_DIVINE_SHIELD] = 'selfbuff1',
		[BENECAST_STRINGS.SELFBUFF_SEAL_OF_FURY] = 'selfbuff2',
		[BENECAST_STRINGS.SELFBUFF_SEAL_OF_JUSTICE] = 'selfbuff3',
		[BENECAST_STRINGS.SELFBUFF_SEAL_OF_LIGHT] = 'selfbuff4',
		[BENECAST_STRINGS.SELFBUFF_SEAL_OF_RIGHTEOUSNESS] = 'selfbuff5',
		[BENECAST_STRINGS.SELFBUFF_SEAL_OF_WISDOM] = 'selfbuff6',
		[BENECAST_STRINGS.SELFBUFF_SEAL_OF_THE_CRUSADER] = 'selfbuff7',
		[BENECAST_STRINGS.SELFBUFF_SEAL_OF_COMMAND] = 'selfbuff8',
		[BENECAST_STRINGS.SELFBUFF_HOLY_SHIELD] = 'selfbuff9',
		[BENECAST_STRINGS.SELFBUFF_DIVINE_FAVOR] = 'selfbuff10',
		[BENECAST_STRINGS.SELFBUFF_RIGHTEOUS_FURY] = 'selfbuff11',
		[BENECAST_STRINGS.BUFF_BLESSING_OF_MIGHT] = 'buff1',
		[BENECAST_STRINGS.BUFF_BLESSING_OF_FREEDOM] = 'buff2',
		[BENECAST_STRINGS.BUFF_BLESSING_OF_LIGHT] = 'buff3',
		[BENECAST_STRINGS.BUFF_BLESSING_OF_SANCTUARY] = 'buff4',
		[BENECAST_STRINGS.BUFF_BLESSING_OF_WISDOM] = 'buff5',
		[BENECAST_STRINGS.BUFF_BLESSING_OF_KINGS] = 'buff6',
		[BENECAST_STRINGS.PARTYBUFF_BLESSING_OF_PROTECTION] = 'buffparty1',
		[BENECAST_STRINGS.PARTYBUFF_BLESSING_OF_SALVATION] = 'buffparty2',
		[BENECAST_STRINGS.PARTYBUFF_BLESSING_OF_SACRIFICE] = 'buffparty3',
		[BENECAST_STRINGS.PARTYBUFF_DIVINE_INTERVENTION] = 'buffparty4',
		[BENECAST_STRINGS.PARTYBUFF_GREATER_BLESSING_OF_MIGHT] = 'buffparty5',
		[BENECAST_STRINGS.PARTYBUFF_GREATER_BLESSING_OF_LIGHT] = 'buffparty6',
		[BENECAST_STRINGS.PARTYBUFF_GREATER_BLESSING_OF_SALVATION] = 'buffparty7',
		[BENECAST_STRINGS.PARTYBUFF_GREATER_BLESSING_OF_WISDOM] = 'buffparty8',
		[BENECAST_STRINGS.PARTYBUFF_GREATER_BLESSING_OF_KINGS] = 'buffparty9',
		[BENECAST_STRINGS.PARTYBUFF_GREATER_BLESSING_OF_SANCTUARY] = 'buffparty10',
	},
	[BENECAST_STRINGS.CLASS_PRIEST] = {
		[BENECAST_STRINGS.RES_RESURRECTION] = 'res',
		[BENECAST_STRINGS.HEAL_LESSER_HEAL] = 'efficient',
		[BENECAST_STRINGS.HEAL_HEAL] = 'efficient',
		[BENECAST_STRINGS.HEAL_GREATER_HEAL] = 'efficient',
		[BENECAST_STRINGS.HEAL_FLASH_HEAL] = 'emergency',
		[BENECAST_STRINGS.HEAL_RENEW] = 'instant',
		[BENECAST_STRINGS.HEAL_PRAYER_OF_HEALING] = 'group',
		[BENECAST_STRINGS.CURE_DISEASE_1] = 'disease',
		[BENECAST_STRINGS.CURE_DISEASE_2] = 'disease',
		[BENECAST_STRINGS.CURE_MAGIC_1] = 'magic',
		[BENECAST_STRINGS.SELFBUFF_FOCUSED_CASTING] = 'selfbuff1',
		[BENECAST_STRINGS.SELFBUFF_INNER_FOCUS] = 'selfbuff2',
		[BENECAST_STRINGS.SELFBUFF_DESPERATE_PRAYER] = 'selfbuff3',
		[BENECAST_STRINGS.SELFBUFF_ELUNES_GRACE] = 'selfbuff4',
		[BENECAST_STRINGS.SELFBUFF_FADE] = 'selfbuff5',
		[BENECAST_STRINGS.SELFBUFF_LEVITATE] = 'selfbuff6',
		[BENECAST_STRINGS.SELFBUFF_SHADOWGUARD] = 'selfbuff7',
		[BENECAST_STRINGS.SELFBUFF_INNER_FIRE] = 'selfbuff8',
		[BENECAST_STRINGS.SELFBUFF_TOUCH_OF_WEAKNESS] = 'selfbuff9',
		[BENECAST_STRINGS.PARTYBUFF_POWER_WORD_SHIELD] = 'buffparty1',
		[BENECAST_STRINGS.BUFF_DIVINE_SPIRIT] = 'buff1',
		[BENECAST_STRINGS.BUFF_POWER_WORD_FORTITUDE] = 'buff2',
		[BENECAST_STRINGS.BUFF_SHADOW_PROTECTION] = 'buff3',
		[BENECAST_STRINGS.BUFF_FEAR_WARD] = 'buff4',
		[BENECAST_STRINGS.GROUPBUFF_PRAYER_OF_FORTITUDE] = 'groupbuff',
		[BENECAST_STRINGS.WEAPONBUFF_FEEDBACK] = 'weaponbuff1',
	},
	[BENECAST_STRINGS.CLASS_ROGUE] = {
		[BENECAST_STRINGS.SELFBUFF_DETECT_TRAPS] = 'selfbuff1',
		[BENECAST_STRINGS.SELFBUFF_BLADE_FLURRY] = 'selfbuff2',
		[BENECAST_STRINGS.SELFBUFF_COLD_BLOOD] = 'selfbuff3',
		[BENECAST_STRINGS.SELFBUFF_ADRENALINE_RUSH] = 'selfbuff4',
		[BENECAST_STRINGS.SELFBUFF_EVASION] = 'selfbuff5',
		[BENECAST_STRINGS.SELFBUFF_PREMEDITATION] = 'selfbuff6',
		[BENECAST_STRINGS.SELFBUFF_PREPARATION] = 'selfbuff7',
		[BENECAST_STRINGS.SELFBUFF_VANISH] = 'selfbuff8',
	},
	[BENECAST_STRINGS.CLASS_SHAMAN] = {
		[BENECAST_STRINGS.RES_ANCESTRAL_SPIRIT] = 'res',
		[BENECAST_STRINGS.HEAL_HEALING_WAVE] = 'efficient',
		[BENECAST_STRINGS.HEAL_LESSER_HEALING_WAVE] = 'emergency',
		[BENECAST_STRINGS.HEAL_CHAIN_HEAL] = 'group',
		[BENECAST_STRINGS.CURE_POISON_1] = 'poison',
		[BENECAST_STRINGS.CURE_DISEASE_1] = 'disease',
		[BENECAST_STRINGS.SELFBUFF_NATURES_SWIFTNESS] = 'selfbuff1',
		[BENECAST_STRINGS.SELFBUFF_LIGHTNING_SHIELD] = 'selfbuff2',
		[BENECAST_STRINGS.BUFF_WATER_BREATHING] = 'buff1',
		[BENECAST_STRINGS.BUFF_WATER_WALKING] = 'buff2',
		[BENECAST_STRINGS.WEAPONBUFF_ROCKBITER_WEAPON] = 'weaponenchant1',
		[BENECAST_STRINGS.WEAPONBUFF_FLAMETONGUE_WEAPON] = 'weaponenchant2',
		[BENECAST_STRINGS.WEAPONBUFF_FROSTBRAND_WEAPON] = 'weaponenchant3',
		[BENECAST_STRINGS.WEAPONBUFF_WINDFURY_WEAPON] = 'weaponenchant4',
	},
	[BENECAST_STRINGS.CLASS_WARLOCK] = {
		[BENECAST_STRINGS.SELFBUFF_DEMON_ARMOR] = 'selfbuff1',
		[BENECAST_STRINGS.SELFBUFF_DEMON_SKIN] = 'selfbuff1',
		[BENECAST_STRINGS.SELFBUFF_DARK_PACT] = 'selfbuff2',
		[BENECAST_STRINGS.SELFBUFF_FEL_DOMINATION] = 'selfbuff3',
		[BENECAST_STRINGS.SELFBUFF_SOUL_LINK] = 'selfbuff4',
		[BENECAST_STRINGS.SELFBUFF_AMPLIFY_CURSE] = 'selfbuff5',
		[BENECAST_STRINGS.SELFBUFF_SHADOW_WARD] = 'selfbuff6',
		[BENECAST_STRINGS.BUFF_DETECT_LESSER_INVIS] = 'buff1',
		[BENECAST_STRINGS.BUFF_DETECT_INVIS] = 'buff1',
		[BENECAST_STRINGS.BUFF_DETECT_GREATER_INVIS] = 'buff1',
		[BENECAST_STRINGS.BUFF_UNENDING_BREATH] = 'buff2',
	},
	[BENECAST_STRINGS.CLASS_WARRIOR] = {
		[BENECAST_STRINGS.SELFBUFF_SHIELD_WALL] = 'selfbuff1',
		[BENECAST_STRINGS.SELFBUFF_RETALIATION] = 'selfbuff2',
		[BENECAST_STRINGS.SELFBUFF_RECKLESSNESS] = 'selfbuff3',
		[BENECAST_STRINGS.SELFBUFF_LAST_STAND] = 'selfbuff4',
		[BENECAST_STRINGS.SELFBUFF_DEATH_WISH] = 'selfbuff5',
		[BENECAST_STRINGS.SELFBUFF_BATTLE_SHOUT] = 'selfbuff6',
		[BENECAST_STRINGS.SELFBUFF_BLOODRAGE] = 'selfbuff7',
		[BENECAST_STRINGS.SELFBUFF_BERSERKER_RAGE] = 'selfbuff8',
		[BENECAST_STRINGS.SELFBUFF_SHIELD_BLOCK] = 'selfbuff9',
		[BENECAST_STRINGS.SELFBUFF_SWEEPING_STRIKES] = 'selfbuff10',
	},
};

-- *****************************************************************************
-- BeneCast Helper functions
-- *****************************************************************************

-- Function to clear the table given as the parameter
local function BeneCast_ClearTable(Table)

	-- Removes every element of the table starting at the end
	while table.getn(Table) > 0 do
		table.remove(Table);
	end
	
end

-- *****************************************************************************
-- BeneCast Option Helper functions
-- *****************************************************************************

-- Function to resize the buttons
local function BeneCast_ResizeButtons()

	local buttonsize = BeneCastConfig.ButtonSize;
	local button, buttonborder, buttoncooldown, buttonfade, frame;
	
	-- Resize the buttons for every unit in _targets
	-- i is the number associated with each unit	
	for i in _targets do		
		-- Resize the buttons based on the saved size for every button in a panel
		for j = 1,BENECAST_MAX_NUMBER_OF_BUTTONS do
			button = getglobal('BeneCastPanel' .. i .. 'Button' .. j);
			-- Set Button height and width to buttonsize
			button:SetHeight(buttonsize);
			button:SetWidth(buttonsize);
			buttonborder = getglobal('BeneCastPanel' .. i .. 'Button' .. j .. 'NormalTexture');
			-- Set the Button border to fit around the button
			buttonborder:SetHeight(buttonsize * 2 - (buttonsize/36) * 10);
			buttonborder:SetWidth(buttonsize * 2 - (buttonsize/36) * 10);
			buttoncooldown = getglobal('BeneCastPanel' .. i .. 'Button' .. j .. 'Cooldown');
			-- Set Cooldown size
			buttoncooldown:SetScale((buttonsize / 36));
			buttoncooldown:SetPoint('CENTER', button, 'CENTER');
			buttonfade = getglobal('BeneCastPanel' .. i .. 'Button' .. j .. 'Fade');
			-- Set Fade size
			buttonfade:SetHeight(buttonsize * 2 - (buttonsize/36) * 10);
			buttonfade:SetWidth(buttonsize * 2 - (buttonsize/36) * 10);
			-- Resize the BeneCastPanelFrame based on the number of buttons shown
			if ( button:IsVisible() ) then
				frame = getglobal('BeneCastPanel' .. i);
				frame:SetWidth(buttonsize * j + 11 + (3 * (j - 1)));
				frame:SetHeight(buttonsize + 11);
			end
		end
	end
	
end


-- Function to load up the saved values in BeneCastConfig
local function BeneCast_LoadOptions()

	-- Load the String Data and saved value into the check buttons
	local button;
	local string;
	local checked;
	for index, value in BeneCastOptionFrameCheckButtons do
		button = getglobal('BeneCastOptionCheckButton'..value.index);
		string = getglobal('BeneCastOptionCheckButton'..value.index..'Text');
		checked = BeneCastConfig[value.cvar];
		button:SetChecked(checked);
		string:SetText(BENECAST_STRINGS[index]);
		button.tooltipText = BENECAST_STRINGS[value.tooltipText];
		-- If this is the tooltip option check to see if it's enabled
		-- If not disable the tooltip name option checkbutton
		if ( value.index == 5 and not checked ) then
			getglobal('BeneCastOptionCheckButton6'):Disable();
		-- If this is the notification option check to see if it's enabled
		-- If not disable all sub notification options
		elseif ( value.index == 8 and not checked ) then
			getglobal('BeneCastOptionCheckButton9'):Disable();
			getglobal('BeneCastOptionCheckButton10'):Disable();
			getglobal('BeneCastOptionCheckButton11'):Disable();
			getglobal('BeneCastOptionCheckButton12'):Disable();
			getglobal('BeneCastOptionCheckButton20'):Disable();
			getglobal('BeneCastOptionCheckButton13'):Disable();
			getglobal('BeneCastOptionCheckButton14'):Disable();
			getglobal('BeneCastOptionCheckButton15'):Disable();
			getglobal('BeneCastOptionCheckButton16'):Disable();
			getglobal('BeneCastOptionCheckButton17'):Disable();
			getglobal('BeneCastOptionCheckButton18'):Disable();
			getglobal('BeneCastOptionCheckButton19'):Disable();
			getglobal('BeneCastOptionCheckButton21'):Disable();
		-- If this is the heal notification option check to see if it's enabled
		-- If not disable all heal max notification
		elseif ( value.index == 16 and not checked ) then
			getglobal('BeneCastOptionCheckButton17'):Disable();
		end
	end

	local slider;
	local string;
	local curvalue;
	-- Load the String Data and saved value into the sliders
	for index, value in BeneCastOptionFrameSliders do
		slider = getglobal('BeneCastOptionSlider'..value.index);
		string = getglobal('BeneCastOptionSlider'..value.index..'Text');
		curvalue = getglobal('BeneCastOptionSlider'..value.index..'Value');
		slider:SetMinMaxValues(value.minValue, value.maxValue);
		slider:SetValueStep(value.valueStep);
		slider:SetValue(BeneCastConfig[value.cvar]);
		curvalue:SetText(BeneCastConfig[value.cvar]);
		string:SetText(BENECAST_STRINGS[value.text]);
		slider.tooltipText = BENECAST_STRINGS[value.tooltipText];
	end
	
	-- Load the user defined chat channel
	if ( BeneCastConfig['UserChannel'] ) then
		getglobal('BeneCastChannelEditBox'):SetText(BeneCastConfig['UserChannel']);
	end

end

-- Loads the data into the spell check buttons
local function BeneCast_LoadSpellConfig()

	-- Position of the current checkbuttons to label
	local i = 1;
	
	-- Name of the spell to make a checkbutton for
	local spellname;
	
	-- Assist checkbutton
	for j = 2,9 do
		getglobal( BeneCastOptionFrameSubframes[j] .. 'Button' .. i .. 'Text' ):SetText(BENECAST_STRINGS.TEXT_ASSIST);
		getglobal( BeneCastOptionFrameSubframes[j] .. 'Button' .. i ):Show();
		getglobal( BeneCastOptionFrameSubframes[j] .. 'Button' .. i ):SetChecked(BeneCastConfig[j]['assist']);
	end
	
	i = i + 1;
	
	-- Load up the spell buttons in the order in _spellsort
	for j in _spellsort do
		-- If this class can cast this spelltype continue
		if _spell_data[_spellsort[j]] then
			-- I think this sets the last spell in a spell line to spellname
			for spell in _spell_types[_class] do
				if _spell_types[_class][spell] == _spellsort[j] then
					spellname = spell;
				end
			end
			-- If this is a selfbuff only add it to the player frame
			if ( string.find(_spellsort[j], 'selfbuff') ) then
				getglobal( BeneCastOptionFrameSubframes[1] .. 'Button' .. i .. 'Text' ):SetText(spellname);
				getglobal( BeneCastOptionFrameSubframes[1] .. 'Button' .. i ):Show();
				getglobal( BeneCastOptionFrameSubframes[1] .. 'Button' .. i ):SetChecked(BeneCastConfig[1][_spellsort[j]]);
			-- Otherwise add it to every spell subframe
			else
				for k = 1,9 do
					getglobal( BeneCastOptionFrameSubframes[k] .. 'Button' .. i .. 'Text' ):SetText(spellname);
					getglobal( BeneCastOptionFrameSubframes[k] .. 'Button' .. i ):Show();
					getglobal( BeneCastOptionFrameSubframes[k] .. 'Button' .. i ):SetChecked(BeneCastConfig[k][_spellsort[j]]);
				end
			end
			-- Increment i if a spell was added to the subframes
			i = i + 1;
		end
	end

end

-- Resets data tables
local function BeneCast_ResetData()

	-- Clear the _spell_data table
	for type in _spell_data do
		BeneCast_ClearTable(_spell_data[type]['spells']);
	end
	
	-- Reload the spell data
	BeneCast_LoadSpellData();
	
	-- Reload the spell checkbuttons
	BeneCast_LoadSpellConfig();
	
end

-- *****************************************************************************
-- BeneCast Frame Attachment Helper functions
-- *****************************************************************************

-- Function to attach BeneCast panels to unit frames
-- Not raid frames, raid frames are handled with BeneCast_AttachRaidPanels
function BeneCast_SnapPanels(arg1)

	local points;
	local panel;
	local parent;
	
	-- Check to make sure the current snap to option can snap to frames
	-- Thanks Jusko!
	if (BeneCast_SnapTo[BeneCastConfig.SnapTo] == nil) then 
		return; 
	end 
	
	-- Snap the BeneCast Panels correct panels based on the value in BeneCastConfig
	if ( BeneCast_SnapTo[BeneCastConfig.SnapTo].AddOn == arg1 ) then
		-- For all the panels in BeneCast_SnapTo for this AddOn do
		for frame in BeneCast_SnapTo[BeneCastConfig.SnapTo] do
			-- Set points to the points for the panel
			points = BeneCast_SnapTo[BeneCastConfig.SnapTo][frame];
			-- So long as this is a BeneCast panel and not the AddOn attribute of the table continue
			if ( frame ~= 'AddOn' and points.frame ~= '' ) then
				-- Set panel to the BeneCastPanel
				panel = getglobal(frame);
				-- Clear the points so it can be moved
				panel:ClearAllPoints();
				-- Set the parent to the frame to attach to
				parent = getglobal(points.frame);
				-- Anchor the panel and set the parent
				panel:SetPoint(points.point, parent, points.relativePoint, points.x, points.y);
				panel:SetParent(parent);
				-- Set up the correct frame strata and level for the panel
				panel:SetFrameStrata(parent:GetFrameStrata());
				panel:SetFrameLevel(parent:GetFrameLevel() + 3);
				-- Set up the correct level for the buttons on the panel
				for x = 1, BENECAST_MAX_NUMBER_OF_BUTTONS do
					getglobal(panel:GetName().. 'Button' .. x):SetFrameLevel(panel:GetFrameLevel() + 1);
					getglobal(panel:GetName().. 'Button' .. x .. 'Cooldown'):SetFrameLevel(panel:GetFrameLevel() + 2);
				end
			end
		end
	end

end

--Function to attach BeneCast panels to raid unit frames
function BeneCast_AttachRaidPanels(AddOn)

	local name, rank, subgroup, level, class, fileName, zone, online, isDead;
	
	local panel, raidframe, raidframetext, nextraidframe;
	
	-- If Raid buttons are not being shown don't spend time attaching frames
	if ( not BeneCastConfig.ShowRaid ) then
		return;
	end
	
	-- Set up all the buttons
	for i = 1, MAX_RAID_MEMBERS do
		name, rank, subgroup, level, class, fileName, zone, online, isDead = GetRaidRosterInfo(i);
		if ( name ) then
			-- Configuration for the standard UI
			if ( AddOn == 'STANDARD' ) then
				-- Attach the panel to the correct frame
				panel = getglobal('BeneCastPanel' .. i + 5);
				for j = 1, BLIZZ_MAX_RAID_PULLOUT_FRAMES do
					for k = 1, BLIZZ_MAX_RAID_PULLOUT_BUTTONS do
						raidframe = getglobal('RaidPullout' .. j .. 'Button' .. k);
						raidframetext = getglobal(raidframe:GetName() .. 'Name'):GetText();
						nextraidframe = nil;
						if ( raidframetext == name and raidframe:IsVisible() ) then
							-- Setup the next raid frame if there it is not the last button in the pullout
							-- And if there's a next frame
							if ( k < MAX_RAID_PULLOUT_BUTTONS and getglobal('RaidPullout' .. j .. 'Button' .. k+1):IsVisible() ) then
								nextraidframe = getglobal('RaidPullout' .. j .. 'Button' .. k+1);
							else
								nextraidframe = nil;
							end
							-- Attach the panel to the frame and set the parent
							panel:SetPoint('TOPLEFT',raidframe,'BOTTOMLEFT');
							panel:SetParent(raidframe);
							-- Set the next raid frame to be underneath the panel
							if ( nextraidframe ) then
								nextraidframe:SetPoint('TOPLEFT',panel,'BOTTOMLEFT');
							end
							-- If the raid unit is not enabled to show a BeneCastPanel move it back
							if ( not BENECAST_RAID_LIST['raid'..i] and nextraidframe ) then
								nextraidframe:SetPoint('TOPLEFT',raidframe,'BOTTOMLEFT',0,-8);
							end
							-- Set up the correct frame strata and level for the panel
							panel:SetFrameStrata(raidframe:GetFrameStrata());
							panel:SetFrameLevel(raidframe:GetFrameLevel() + 3);
							-- Set up the correct level for the buttons on the panel
							for x = 1, BENECAST_MAX_NUMBER_OF_BUTTONS do
								getglobal(panel:GetName().. 'Button' .. x):SetFrameLevel(panel:GetFrameLevel() + 1);
								getglobal(panel:GetName().. 'Button' .. x .. 'Cooldown'):SetFrameLevel(panel:GetFrameLevel() + 2);
							end
						end
					end
				end
			-- Configuration is for CT_RAID
			elseif ( AddOn == 'CT_RAID' ) then
				-- Attach the panel to the correct frame
				panel = getglobal('BeneCastPanel' .. i + 5);
				for j = 1, MAX_RAID_MEMBERS do
					raidframe = getglobal('CT_RAMember' .. j);
					raidframetext = getglobal(raidframe:GetName() .. 'Name'):GetText();
					nextraidframe = nil;
					if ( raidframetext == name and raidframe:IsVisible() ) then
						-- Setup the next raid frame if there it is not the last button in the pullout
						if ( raidframe.next ) then
							nextraidframe = raidframe.next;
						else
							nextraidframe = nil;
						end
						-- Attach the panel to the frame and set the parent
						panel:SetPoint('TOPLEFT',raidframe,'BOTTOMLEFT');
						panel:SetParent(raidframe);
						-- Set the next raid frame to be underneath the panel
						if ( nextraidframe ) then
							nextraidframe:SetPoint('TOPLEFT',panel,'BOTTOMLEFT');
						end
						-- If the raid unit is not enabled to show a BeneCastPanel move it back
						if ( not BENECAST_RAID_LIST['raid'..i] and nextraidframe ) then
							nextraidframe:SetPoint('TOPLEFT',raidframe,'BOTTOMLEFT',0,0);
						end
						-- Set up the correct frame strata and level for the panel
						panel:SetFrameStrata(raidframe:GetFrameStrata());
						panel:SetFrameLevel(raidframe:GetFrameLevel() + 3);
						-- Set up the correct level for the buttons on the panel
						for x = 1, BENECAST_MAX_NUMBER_OF_BUTTONS do
							getglobal(panel:GetName().. 'Button' .. x):SetFrameLevel(panel:GetFrameLevel() + 1);
							getglobal(panel:GetName().. 'Button' .. x .. 'Cooldown'):SetFrameLevel(panel:GetFrameLevel() + 2);
						end
					end
				end
			end
		end
	end

end

-- *****************************************************************************
-- BeneCast Raid Helper functions
-- *****************************************************************************

local function BeneCast_ClearRaid()

	-- Reset all the buttons
	for i = 1, MAX_RAID_MEMBERS do
		getglobal('BeneCastRaidButton' .. i):SetChecked('false');
		getglobal('BeneCastRaidButton' .. i):Disable();
		getglobal('BeneCastRaidButton' .. i .. 'Empty'):SetText(BENECAST_STRINGS.TEXT_EMPTY);
		getglobal('BeneCastRaidButton' .. i .. 'Empty'):SetTextColor(GRAY_FONT_COLOR.r, GRAY_FONT_COLOR.g, GRAY_FONT_COLOR.b);
		getglobal('BeneCastRaidButton' .. i .. 'Name'):SetText('');
		getglobal('BeneCastRaidButton' .. i .. 'Level'):SetText('');
		getglobal('BeneCastRaidButton' .. i .. 'Class'):SetText('');
	end

end

local function BeneCast_UpdateRaid()

	local name, rank, subgroup, level, class, fileName, zone, online, isDead;
	
	-- Reset the Subgroups table
	for i = 1, NUM_RAID_GROUPS do
		BENECAST_RAID_SUBGROUPS[i] = 0;
	end
	
	-- Reset all the buttons
	BeneCast_ClearRaid();
	
	-- Set up all the buttons
	for i = 1, MAX_RAID_MEMBERS do
		name, rank, subgroup, level, class, fileName, zone, online, isDead = GetRaidRosterInfo(i);
		BENECAST_RAID_SUBGROUPS[subgroup] = BENECAST_RAID_SUBGROUPS[subgroup]+1
		if ( name ) then
			if ( BENECAST_RAID_LIST['raid'..i] ) then
				getglobal('BeneCastRaidButton' .. (subgroup-1)*MEMBERS_PER_RAID_GROUP + BENECAST_RAID_SUBGROUPS[subgroup]):SetChecked('true');
			end
			getglobal('BeneCastRaidButton' .. (subgroup-1)*MEMBERS_PER_RAID_GROUP + BENECAST_RAID_SUBGROUPS[subgroup]):Enable();
			getglobal('BeneCastRaidButton' .. (subgroup-1)*MEMBERS_PER_RAID_GROUP + BENECAST_RAID_SUBGROUPS[subgroup] .. 'Empty'):SetText(nil);
			getglobal('BeneCastRaidButton' .. (subgroup-1)*MEMBERS_PER_RAID_GROUP + BENECAST_RAID_SUBGROUPS[subgroup] .. 'Empty'):SetTextColor(GRAY_FONT_COLOR.r, GRAY_FONT_COLOR.g, GRAY_FONT_COLOR.b);
			getglobal('BeneCastRaidButton' .. (subgroup-1)*MEMBERS_PER_RAID_GROUP + BENECAST_RAID_SUBGROUPS[subgroup] .. 'Name'):SetText(name);
			getglobal('BeneCastRaidButton' .. (subgroup-1)*MEMBERS_PER_RAID_GROUP + BENECAST_RAID_SUBGROUPS[subgroup] .. 'Name'):SetTextColor(RAID_CLASS_COLORS[fileName].r, RAID_CLASS_COLORS[fileName].g, RAID_CLASS_COLORS[fileName].b);
			getglobal('BeneCastRaidButton' .. (subgroup-1)*MEMBERS_PER_RAID_GROUP + BENECAST_RAID_SUBGROUPS[subgroup] .. 'Level'):SetText(level);
			getglobal('BeneCastRaidButton' .. (subgroup-1)*MEMBERS_PER_RAID_GROUP + BENECAST_RAID_SUBGROUPS[subgroup] .. 'Level'):SetTextColor(RAID_CLASS_COLORS[fileName].r, RAID_CLASS_COLORS[fileName].g, RAID_CLASS_COLORS[fileName].b);
			getglobal('BeneCastRaidButton' .. (subgroup-1)*MEMBERS_PER_RAID_GROUP + BENECAST_RAID_SUBGROUPS[subgroup] .. 'Class'):SetText(class);
			getglobal('BeneCastRaidButton' .. (subgroup-1)*MEMBERS_PER_RAID_GROUP + BENECAST_RAID_SUBGROUPS[subgroup] .. 'Class'):SetTextColor(RAID_CLASS_COLORS[fileName].r, RAID_CLASS_COLORS[fileName].g, RAID_CLASS_COLORS[fileName].b);		
		end
	end
	
	-- Attach the Raid Panels to the correct frames
	BeneCast_AttachRaidPanels(BeneCastConfig.RaidSnapTo)
	
end

-- *****************************************************************************
-- BeneCast Unit Helper functions
-- *****************************************************************************

-- Function to determine if a unit is in your raid group
local function BeneCast_UnitInRaid(unit)

	for i = 1, 40 do
		if (UnitIsUnit('raid' .. i, unit)) then
			return true;
		end
	end
	return nil;
	
end

-- Function to determine if a unitid is 'raid'..x
local function BeneCast_UnitIsRaidID(unit)

	for i = 1, 40 do
		if ( 'raid' .. i == unit ) then
			return true;
		end
	end
	return nil;
	
end

-- Function to determine if a unitid is x..'pet'..y
local function BeneCast_UnitIsPetID(unit)

	return string.find(unit, 'pet');
	
end

-- Function to determine if a unitid is 'raidpet'..x
local function BeneCast_UnitIsRaidPetID(unit)

	return string.find(unit, 'raidpet');
	
end


-- Function to determine if a unit is a party pet
local function BeneCast_UnitIsPartyPet(unit)

	for i = 1, 4 do
		if (UnitIsUnit('partypet' .. i, unit)) then
			return true;
		end
	end
	return nil;
	
end

-- Function to determine if a unit is a raid pet
local function BeneCast_UnitIsRaidPet(unit)

	for i = 1, 40 do
		if (UnitIsUnit('raidpet' .. i, unit)) then
			return true;
		end
	end
	return nil;
	
end

-- Function to determine the BeneCastPanel number of a UnitId
local function BeneCast_UnitID(unit)

	for i in _targets do
		if unit == _targets[i] then
			return i;
		end
	end
	return nil;

end

-- *****************************************************************************
-- BeneCast tooltip parsing functions
-- *****************************************************************************

-- Functions for parsing healing info from tooltip. Thanks Danboo!
local function parse_integer(text, index)
	if (not text) then
	return 0;
	end 
	local i = 1;
	-- Replace ',' with '.', needed for German client. Thanks Auric!
	text = string.gsub(text, ',', '.');
	for integer in string.gfind(text, '%d+%.?%d*') do
		if ( i == index ) then
			return tonumber(integer);
		end
		i = i + 1;
	end

end

-- Pull the first integer
local function pi_first(text)
	return parse_integer(text, 1);
end

-- Pull the second integer
local function pi_second(text)
	return parse_integer(text, 2);
end

-- Pull the second integer
local function pi_third(text)
	return parse_integer(text, 3);
end

-- Pull the second integer
local function pi_fourth(text)
	return parse_integer(text, 4);
end

-- Mapping of healing spell lines to heal amount parsers. Thanks Danboo!
-- Spell parsing is slightly different in the German client
local _spell_parsers = {}
if ( GetLocale() == deDE ) then
	_spell_parsers = {
		[BENECAST_STRINGS.CLASS_SHAMAN] = {
				[BENECAST_STRINGS.HEAL_HEALING_WAVE] = pi_second,
				[BENECAST_STRINGS.HEAL_LESSER_HEALING_WAVE] = pi_second,
				[BENECAST_STRINGS.HEAL_CHAIN_HEAL] = pi_second,
		},
		[BENECAST_STRINGS.CLASS_PRIEST] = {
				[BENECAST_STRINGS.HEAL_LESSER_HEAL] = pi_second,
				[BENECAST_STRINGS.HEAL_HEAL] = pi_second,
				[BENECAST_STRINGS.HEAL_GREATER_HEAL] = pi_second,
				[BENECAST_STRINGS.HEAL_FLASH_HEAL] = pi_second,
				[BENECAST_STRINGS.HEAL_RENEW] = pi_second,
				[BENECAST_STRINGS.HEAL_PRAYER_OF_HEALING] = pi_second,
		},
		[BENECAST_STRINGS.CLASS_DRUID] = {
				[BENECAST_STRINGS.HEAL_HEALING_TOUCH] = pi_second,
				[BENECAST_STRINGS.HEAL_REJUVENATION] = pi_second,
				[BENECAST_STRINGS.HEAL_REGROWTH] = pi_second,
				[BENECAST_STRINGS.HEAL_TRANQUILITY] = pi_third,
		},
		[BENECAST_STRINGS.CLASS_PALADIN] = {
				[BENECAST_STRINGS.HEAL_HOLY_LIGHT] = pi_second,
				[BENECAST_STRINGS.HEAL_FLASH_OF_LIGHT] = pi_second,
				[BENECAST_STRINGS.HEAL_LAY_ON_HANDS] = pi_second,
				[BENECAST_STRINGS.HEAL_HOLY_SHOCK] = pi_fourth,
		},
	};
else
	_spell_parsers = {
		[BENECAST_STRINGS.CLASS_SHAMAN] = {
			[BENECAST_STRINGS.HEAL_HEALING_WAVE] = pi_second,
			[BENECAST_STRINGS.HEAL_LESSER_HEALING_WAVE] = pi_second,
			[BENECAST_STRINGS.HEAL_CHAIN_HEAL] = pi_second,
		},
		[BENECAST_STRINGS.CLASS_PRIEST] = {
			[BENECAST_STRINGS.HEAL_LESSER_HEAL] = pi_second,
			[BENECAST_STRINGS.HEAL_HEAL] = pi_second,
			[BENECAST_STRINGS.HEAL_GREATER_HEAL] = pi_second,
			[BENECAST_STRINGS.HEAL_FLASH_HEAL] = pi_second,
			[BENECAST_STRINGS.HEAL_RENEW] = pi_first,
			[BENECAST_STRINGS.HEAL_PRAYER_OF_HEALING] = pi_second,
		},
		[BENECAST_STRINGS.CLASS_DRUID] = {
			[BENECAST_STRINGS.HEAL_HEALING_TOUCH] = pi_second,
			[BENECAST_STRINGS.HEAL_REJUVENATION] = pi_first,
			[BENECAST_STRINGS.HEAL_REGROWTH] = pi_second,
			[BENECAST_STRINGS.HEAL_TRANQUILITY] = pi_first,
		},
		[BENECAST_STRINGS.CLASS_PALADIN] = {
			[BENECAST_STRINGS.HEAL_HOLY_LIGHT] = pi_second,
			[BENECAST_STRINGS.HEAL_FLASH_OF_LIGHT] = pi_second,
			[BENECAST_STRINGS.HEAL_LAY_ON_HANDS] = pi_second,
			[BENECAST_STRINGS.HEAL_HOLY_SHOCK] = pi_fourth,
		},
	};
end

-- *****************************************************************************
-- BeneCast Spell Loading function
-- *****************************************************************************

-- Fill in the _spell_data table by iterating over spells in spellbook
-- and parsing tool tips for info. Thanks Danboo!
function BeneCast_LoadSpellData()
	
	-- Iterator to move through the spellbook
	local i = 1;

	while true do
		
		-- Spell at position i in the player's spellbook
		local spell_name, spell_rank = GetSpellName(i, SpellBookFrame.bookType);
		
		-- If there is no spell at i, break out of the loop
		if not spell_name then
			do break end
		end
		
		-- Load up the spell_type based on the spell name and player class
		local spell_type = _spell_types[_class][spell_name];

		-- If the spell_type exists do stuff
		if ( spell_type ) then
						
			-- Set BeneCast_Tooltip to the spell info
			-- Be sure to clear it first
			BeneCast_TooltipTextLeft2:SetText('');
			BeneCast_TooltipTextLeft3:SetText('');
			BeneCast_TooltipTextLeft4:SetText('');
			BeneCast_Tooltip:SetSpell(i, SpellBookFrame.bookType);
			
			local _,_,mana_cost;
			-- If this spell has a set mana cost set mana_cost to it
			if ( BeneCast_TooltipTextLeft2:GetText() and string.find(BeneCast_TooltipTextLeft2:GetText(), '^(%d+) Mana') ) then 
				_,_,mana_cost = string.find(BeneCast_TooltipTextLeft2:GetText(), '^(%d+) Mana');
			-- If this spell does not have a set mana cost, but does have a rank set mana_cost to the rank
			-- This is to ensure correct spell ordering
			elseif ( spell_rank ~= '' ) then
				mana_cost = string.find(spell_rank, '%d+');
			-- If this spell does not have a set mana cost or rank set mana_cost to zero
			-- I'm guessing there is no need to order a spell type if it has no ranks
			else
				mana_cost = 0;
			end
			
			local heal_amount;
			-- If the player class is Paladin and the spell is Lay on Hands set heal_amount to zero
			-- This is to ensure that the highest rank is always used when used
			if ( _class == BENECAST_STRINGS.CLASS_PALADIN and spell_type == 'instant' ) then
				heal_amount = 0
			-- If the spell_type is a heal spell then set heal_amount to the amount this spell heals
			elseif ( spell_type == 'efficient' or spell_type == 'emergency' or spell_type == 'instant' or spell_type == 'group' ) then 
				if ( BeneCast_TooltipTextLeft4:GetText() ) then
					heal_amount = _spell_parsers[_class][spell_name](BeneCast_TooltipTextLeft4:GetText());
				else
					heal_amount = _spell_parsers[_class][spell_name](BeneCast_TooltipTextLeft3:GetText());
				end
			end
			
			-- If there isn't anything in _spell_data[spell_type] yet then create a table for it
			if ( not _spell_data[spell_type] ) then
				_spell_data[spell_type] = {};
				_spell_data[spell_type]['spells'] = {};
			end
			
			local entry;
			-- If the entry type is a heal spell be sure to include rank and max_heal information
			if ( spell_type == 'efficient' or spell_type == 'emergency' or spell_type == 'instant' or spell_type == 'group' ) then
				entry = { name = (spell_name),
					rank = (spell_rank),
					max_heal = tonumber(heal_amount),
					mana = tonumber(mana_cost),
					id = i };
			-- If the entry is a cure you don't need the rank
			elseif ( spelltype == 'poison' or spelltype == 'disease' or spelltype == 'magic' or spelltype == 'curse' ) then
				entry = {name = (spell_name),
					mana = tonumber(mana_cost),
					id = i };
			-- Otherwise it's a buff and include the spell rank if you can
			elseif ( spell_rank ~= '' ) then
				entry = {name = (spell_name),
					rank = (spell_rank),
					mana = tonumber(mana_cost),
					id = i };
			-- If there is no rank, don't put a spell rank into the entry
			else
				entry = {name = (spell_name),
					mana = tonumber(mana_cost),
					id = i };
			end
			
			-- Insert the entry into the appropriate table
			table.insert(_spell_data[spell_type]['spells'], entry);

		end

		i = i + 1;

	end
	
	-- Sort the spells by mana cost within each spell type
	for type in _spell_data do
		-- Purify, Cleanse, Abolish Poison/Disease and Cure Poison/Disease are not sorted by mana but by name
		if ( type == 'poison' or type == 'disease' ) then
			table.sort(_spell_data[type]['spells'], function(a,b) return a.name > b.name end);
		else
			table.sort(_spell_data[type]['spells'], function(a,b) return a.mana < b.mana end);
		end
	end
	
	-- If the spell_type does not have a texture associated with it set it
	for type in _spell_data do
		local texture;
		-- Make sure the texture is for the last entry in the table
		for i in _spell_data[type]['spells'] do
			texture = GetSpellTexture(_spell_data[type]['spells'][i].id, SpellBookFrame.bookType);
		end
		_spell_data[type].texture = texture;
		_spell_data[type].type = type;
	end
	
end

-- *****************************************************************************
-- BeneCast spell choose functions
-- *****************************************************************************

-- Interpolates the health of the given unit (used for those outside the party). Thanks Danboo!
-- Returns an estimated health for a target outside the party
function BeneCast_UnitInterpolateHealth(unit, health_percent)
	
	local class      = UnitClass(unit);
	local level      = UnitLevel(unit);
	local slope      = ( _class_health[class]['60'] - _class_health[class]['1'] ) / 59;
	local health_max = level * slope + _class_health[class]['1'];
	local health     = health_max * health_percent / 100;
	return floor(health), floor(health_max);
	
end

-- Function to Choose the appropriate spell to cast. Thanks Danboo!
-- Returns the spell to be cast
function BeneCast_ChooseSpell(type, target, forcemax, forceoverheal)

	if ( type == 'assist' ) then
		return;
	end

	local mana = UnitMana('player');
	-- If the clearcasting state is true treat mana as full
	if ( _clearcasting ) then
		mana = UnitManaMax('player');
	end
	local maxhealth = 0;
	local health = 0;
	-- If the type is group then caculate the total party damage percentage
	if ( type == 'group' ) then
		for i = 1,5 do
			maxhealth = maxhealth + UnitHealthMax(_targets[i]);
			health = health + UnitHealth(_targets[i]);
		end
		maxhealth = maxhealth * _threshold;
	-- Else calculate the damage percentage of the target
	else
		maxhealth = UnitHealthMax(target) * _threshold;
		health = UnitHealth(target);
	end
	-- If the target is not in the party/raid use UnitInterpolateHealth to determine health
	if ( not UnitInParty(target) and not BeneCast_UnitInRaid(target) and not BeneCast_UnitIsPartyPet(target) and not BeneCast_UnitIsRaidPet(target) ) then
		health, maxhealth = BeneCast_UnitInterpolateHealth(target, health);
	end
	-- If forcemax is true, cast as if the target had 0 health
	if ( forcemax ) then
		health = 0;
	end
	
	-- overhealamount is the number of ranks to overheal from the targets current hp
	local overhealamount = 0;
	if ( forceoverheal ) then
		overhealamount = BeneCastConfig.OverhealSlider;
	end
	
	local spell;
	
	-- If the type is a heal then use the loop that looks at max_heal
	if ( type == 'efficient' or type == 'emergency' or type == 'instant' or type == 'group' ) then
		for i in _spell_data[type]['spells'] do
			-- Check to see if the player has enough mana and the targets level is enough to be targeted
			-- Level only matters on HoTs
			if ( _spell_data[type]['spells'][i].mana <= mana and UnitLevel(target) >= (_spell_level[_class][type][i] - 10) ) then
				-- Set spell to the current best cast
				spell = _spell_data[type]['spells'][i];
				-- If this heal bring the target over the threshold break out of the loop
				if ( _spell_data[type]['spells'][i].max_heal + health >= maxhealth ) then
					if ( overhealamount == 0 ) then
						do break end
					end
					overhealamount = overhealamount - 1;
				end
			-- Set the spell to the lowest possible cast if spell is not set. Used to announce 'Out of Mana' and for tooltips
			elseif ( not spell ) then
				spell = _spell_data[type]['spells'][i];
			-- Break out of loop early if not enough mana or target is too low of level or too low level for the HoT
			else
				do break end
			end
		end
	-- If type is a cure effect then always cast the best cure, since they all take the same mana
	elseif ( type == 'poison' or type == 'disease' or type == 'magic' or type == 'curse' ) then
		for i in _spell_data[type]['spells'] do
			-- Check to see if the player has enough mana. Pretty sure there's no level limit for this
			if ( _spell_data[type]['spells'][i].mana <= mana ) then
				spell = _spell_data[type]['spells'][i];
			-- Set the spell to the lowest possible cast if spell is not set. Used to announce 'Out of Mana' and for tooltips
			elseif ( not spell ) then
				spell = _spell_data[type]['spells'][i];
			-- Break out of loop early if not enough mana
			else
				do break end
			end
		end
	-- Else the spell being cast is a buff, cast the highest one possible
	else
		for i in _spell_data[type]['spells'] do
			-- Check to see if the targets level is enough to be targeted. Mana is not an issue when buffing, always cast the max possible
			if ( UnitLevel(target) >= (_spell_level[_class][type][i] - 10) ) then
				spell = _spell_data[type]['spells'][i];
			-- Set the spell to the lowest possible cast if spell is not set. Used to announce 'Out of Mana' and for tooltips
			elseif ( not spell ) then
				spell = _spell_data[type]['spells'][i];
			-- Break out of loop early if the target is too low of level. Pretty sure level limits buff strength
			else
				do break end
			end
		end
	end
	
	-- Return the spell to be cast
	return spell;

end

-- *****************************************************************************
-- BeneCast Slash Command Handler
-- *****************************************************************************

-- Slash command handler
-- The only slash command is to show the config GUI
function BeneCast_SlashCommandHandler(msg)

	BeneCastOptionFrameToggle('LeftButton');
	
end

-- *****************************************************************************
-- BeneCast Key Binding functions
-- *****************************************************************************

-- Function to toggle showing of all buffs
function BeneCast_ToggleBuffs()
	
	-- Toggle showallbuffs
	BeneCastConfig.ShowAllBuffs = not BeneCastConfig.ShowAllBuffs;
	
	for i in _targets do
		BeneCast_UpdateButtons(i);
	end

end

-- Function to cast spells through a binding
function BeneCast_BindingCast(id)

	-- The default target for the spell is the player
	local target = 'player';
	
	-- If there is a target and it's friendly choose the target as the target for the spell
	if ( UnitExists('target') and not UnitCanAttack('player', 'target') ) then
		target = 'target';
	end
	
	local spell;
	local spelltype;
	
	if ( target == 'player' ) then
		if ( not getglobal('BeneCastPanel1Button' .. id):IsVisible() ) then
			return;
		end
		spelltype = getglobal('BeneCastPanel1Button' .. id).spelltype;
	else
		if ( not getglobal('BeneCastPanel91Button' .. id):IsVisible() ) then
			return;
		end
		spelltype = getglobal('BeneCastPanel91Button' .. id).spelltype;
	end
	
	-- If a spell is already targeting stop targeting
	if ( SpellIsTargeting() ) then
		SpellStopTargeting();
	end
	
	-- Always take care of items on cursor with left clicks
	if ( button == 'LeftButton' and CursorHasItem() ) then
		if ( unit == 'player' ) then
			AutoEquipCursorItem();
		else
			DropItemOnUnit(target);
		end
		return;
	end

	-- Determine if the max heal has been forced with the shift key
	local forcemax = IsShiftKeyDown();
	
	-- Determine if the overheal has been forced with the ctrl key
	local forceoverheal = IsControlKeyDown();
	
	-- If maxheal is not true then reverse the functionality of forcemax
	if ( not BeneCastConfig.DmgBasedHeal ) then
		forcemax = not forcemax;
	end
	
	-- If overheal is true, then reverse the functionality of overheal	
	if ( BeneCastConfig.Overheal ) then
		forceoverheal = not forceoverheal;
	end

	spell = BeneCast_ChooseSpell(spelltype, target, forcemax, forceoverheal);
	
	-- In case you're targeting an enemy and cast Dispel Magic
	local retarget_enemy = nil;
	
	-- Try casting the spell if it exists
	if ( spell ) then
		-- Dispel Magic can be used offensively, always target the intended target
		if ( _class == BENECAST_STRINGS.CLASS_PRIEST and spelltype == 'magic' ) then
			retarget_enemy = UnitCanAttack('player', 'target');
			TargetUnit(target);
		end
		-- Cast the spell
		CastSpell(spell.id, SpellBookFrame.bookType);
		_spellbeingcast = spell;
		_spelltarget = target;
		_spellismax = forcemax;
		SpellTargetUnit(target);
	end
	
	-- Re-target the last enemy if you changed targets due to Dispel Magic
	if ( retarget_enemy ) then
		TargetLastEnemy();
	end
	
end

-- *****************************************************************************
-- BeneCast Party Notification functions
-- *****************************************************************************

-- Function for Party Notification
-- Modified SpeakSpell by Danboo. Thanks Danboo!
-- Returns true if a notification message was sent
local function BeneCast_PartyNotify(target, spell, ismax, duration)

	-- If there is no duration, give a duration of 0
	if ( not duration ) then
		duration = 0;
	end
		
	if ( not BeneCastConfig.Notification ) then
		return nil;
	end
	
	-- Don't show time if notifytime is nil or duration is 0
	local time = nil;
	if ( duration > 0 and BeneCastConfig.NotifyTime ) then
		time = true;
	end
	
	-- Don't show rank if spell.rank is nil
	local rank = nil;
	if ( spell.rank and spell.rank ~= 'Rank 0' ) then
		rank = true;
	else
		spell.rank = 'Rank 0';
	end
	
	-- If the spell is targeting the player and not allowing self casts then return
	if ( not BeneCastConfig.SelfCasts and target == 'player' ) then
		return nil;
	end
	
	-- Figure out the spelltype for display purposes
	local type = _spell_types[_class][spell.name];
	
	-- If the type is a heal and if notification for heals is false then return
	if ( type == 'efficient' or type == 'emergency' or type == 'instant' or type == 'group' ) then
		if ( not BeneCastConfig.NotifyHeal ) then
			return nil;
		-- else if notification for only max heals and not a max heal then return
		elseif ( BeneCastConfig.NotifyMaxHeal and not ismax ) then
			return nil;
		end
	-- If the type is a cure and if notification for cures is false then return
	elseif ( type == 'poison' or type == 'disease' or type == 'magic' or type == 'curse' ) then
		if ( not BeneCastConfig.NotifyCure ) then
			return nil;
		end
	-- If the type is a res and if notification for res' is false then return
	elseif ( type == 'res' ) then
		if ( not BeneCastConfig.NotifyRes ) then
			return nil;
		end
	-- The type is a buff and if notification for buffs is false then return
	else
		if ( not BeneCastConfig.NotifyBuff ) then
			return nil;
		end
	end

	duration = string.format('%.1f', duration / 1000);

	local message_text;
	
	-- Set the message text based on what options are checked
	if ( rank and BeneCastConfig.NotifyRank and time ) then
		message_text = BENECAST_STRINGS.NOTIFY_ALL;
	elseif ( rank and BeneCastConfig.NotifyRank ) then
		message_text = BENECAST_STRINGS.NOTIFY_RANK;
	elseif ( time ) then
		message_text = BENECAST_STRINGS.NOTIFY_TIME;
	else
		message_text = BENECAST_STRINGS.NOTIFY;
	end
	
	message_text = string.gsub(message_text, '%%s', spell.name);
	-- If notifying the target, address them as 'you'
	if ( BeneCastConfig.TargetNotify ) then
		message_text = string.gsub(message_text, '%%t', BENECAST_STRINGS.NOTIFY_YOU);
	-- In laggy environments if the player's target was deselected while casting it does not have a name
	elseif ( UnitExists(target) ) then
		message_text = string.gsub(message_text, '%%t', UnitName(target));
	else
		message_text = string.gsub(message_text, '%%t', BENECAST_STRINGS.NOTIFY_UNKNOWN_ENTITY);
	end
	message_text = string.gsub(message_text, '%%d', duration);
	message_text = string.gsub(message_text, '%%r', spell.rank);

	-- Send the message on the appropriate channel
	if ( BeneCastConfig.SelfNotify and DEFAULT_CHAT_FRAME ) then
		DEFAULT_CHAT_FRAME:AddMessage(message_text);
	elseif ( BeneCastConfig.PartyNotify and UnitExists('party1') ) then
		SendChatMessage(message_text, 'PARTY');
	elseif ( BeneCastConfig.RaidNotify and UnitExists('raid1') ) then
		SendChatMessage(message_text, 'RAID');
	elseif ( BeneCastConfig.TargetNotify and target ~= 'player' and UnitIsPlayer(target) ) then
		SendChatMessage(message_text, 'WHISPER', nil, UnitName(target));
	elseif ( BeneCastConfig.UserNotify ) then
		SendChatMessage(message_text, 'CHANNEL', nil, GetChannelName(BeneCastConfig.UserChannel));
	else
		return nil;
	end
	
	-- Message sent, return true
	return true;

end

-- Function for Party Notification on failures and interruptions
-- Modified SpeakSpell by Danboo. Thanks Danboo!
local function BeneCast_PartyNotifyFail(event)

	local message_text;
	
	-- Set the message text based on what event called PartyNotifyFail
	-- SPELLCAST_CANCELED is really SPELLCAST_INTERRUPTED if the interruption was caused by the player
	-- To know if interruptions are caused by the player, SPELLCAST_STOP is called just before SPELLCAST_INTERRUPTED
	if event == 'SPELLCAST_FAILED' then
		message_text = BENECAST_STRINGS.NOTIFY_FAILED;
	elseif event == 'SPELLCAST_INTERRUPTED' then
		message_text = BENECAST_STRINGS.NOTIFY_INTERRUPTED;
	elseif event == 'SPELLCAST_CANCELED' then
		message_text = BENECAST_STRINGS.NOTIFY_CANCELED;
	end

	-- Send the message on the appropriate channel
	if ( BeneCastConfig.SelfNotify and DEFAULT_CHAT_FRAME ) then
		DEFAULT_CHAT_FRAME:AddMessage(message_text);
	elseif ( BeneCastConfig.PartyNotify and UnitExists('party1') ) then
		SendChatMessage(message_text, 'PARTY');
	elseif ( BeneCastConfig.RaidNotify and UnitExists('party1') ) then
		SendChatMessage(message_text, 'RAID');
	elseif ( BeneCastConfig.TargetNotify and target ~= 'player' and UnitIsPlayer(target) ) then
		SendChatMessage(message_text, 'WHISPER', nil, UnitName(target));
	elseif ( BeneCastConfig.UserNotify ) then
		SendChatMessage(message_text, 'CHANNEL', nil, GetChannelName(BeneCastConfig.UserChannel));
	end

end

-- *****************************************************************************
-- BeneCastButton functions
-- *****************************************************************************

-- Do the Flash animation for a button
function BeneCastButton_OnUpdate()

	local buffAlphaValue;
	if ( this.flashing ) then
		if ( BuffFrameFlashState == 1 ) then
			buffAlphaValue = (BUFF_FLASH_TIME_ON - BuffFrameFlashTime) / BUFF_FLASH_TIME_ON;
			buffAlphaValue = buffAlphaValue * (1 - BUFF_MIN_ALPHA) + BUFF_MIN_ALPHA;
		else
			buffAlphaValue = BuffFrameFlashTime / BUFF_FLASH_TIME_ON;
			buffAlphaValue = (buffAlphaValue * (1 - BUFF_MIN_ALPHA)) + BUFF_MIN_ALPHA;
			this:SetAlpha(BuffFrameFlashTime / BUFF_FLASH_TIME_ON);
		end
		this:SetAlpha(buffAlphaValue);
	else
		this:SetAlpha(1);
	end
	
end

-- Cast a spell when a BeneCastButton is clicked on
function BeneCastButton_OnClick()

	local target = _targets[this:GetParent():GetID()];
	-- In case you switch targets because you're targeting a friendly
	local retarget_friend = nil;
	
	if ( not UnitExists(target) ) then
		return;
	end
	
	local spell;
	local spelltype = this.spelltype;
	
	if ( this.spelltype == 'assist' ) then
		AssistUnit(target);
		return;
	end
	
	-- Target the intended target if the current target is friendly but not the intended target
	if ( not UnitCanAttack('player', 'target') and not UnitIsUnit('target', target) and not string.find(spelltype, 'selfbuff') and UnitExists('target') ) then
		retarget_friend = true;			
		TargetUnit(target);
	end
	
	-- If a spell is already targeting stop targeting
	if ( SpellIsTargeting() ) then
		SpellStopTargeting();
	end
	
	-- Always take care of items on cursor with left clicks
	if ( button == 'LeftButton' and CursorHasItem() ) then
		if ( unit == 'player' ) then
			AutoEquipCursorItem();
		else
			DropItemOnUnit(target);
		end
		return;
	end

	-- Determine if the max heal has been forced with the shift key
	local forcemax = IsShiftKeyDown();
	
	-- Determine if the overheal has been forced with the ctrl key
	local forceoverheal = IsControlKeyDown();
	
	-- If maxheal is not true then reverse the functionality of forcemax
	if ( not BeneCastConfig.DmgBasedHeal ) then
		forcemax = not forcemax;
	end
	
	-- If overheal is true, then reverse the functionality of overheal	
	if ( BeneCastConfig.Overheal ) then
		forceoverheal = not forceoverheal;
	end

	-- Cast Nature's Swiftness before casting the a Druid or Shaman heal if the Alt key is held down
	if ( IsAltKeyDown() and ( _class == BENECAST_STRINGS.CLASS_DRUID or _class == BENECAST_STRINGS.CLASS_SHAMAN ) and ( spelltype == 'efficient' or spelltype == 'emergency' ) and _spell_data['selfbuff1'] ) then
		local natureswiftness = _spell_data['selfbuff1']['spells'][1];
		CastSpell(natureswiftness.id, SpellBookFrame.bookType);
		SpellStopCasting();
		forcemax = true;
	end

	spell = BeneCast_ChooseSpell(spelltype, target, forcemax, forceoverheal);
	
	-- In case you're targeting an enemy and cast make sure that you'll cast on the intended target
	local retarget_enemy = nil;
	
	-- Try casting the spell if it exists
	if ( spell ) then
		-- Dispel Magic can be used offensively, always target the intended target
		if ( _class == BENECAST_STRINGS.CLASS_PRIEST and spelltype == 'magic' and UnitCanAttack('player', 'target') ) then
			retarget_enemy = true;
			TargetUnit(target);
		end
		-- Cast the spell
		CastSpell(spell.id, SpellBookFrame.bookType);
		_spellbeingcast = spell;
		_spelltarget = target;
		_spellismax = forcemax;
		SpellTargetUnit(target);
	end
	
	-- Re-target the last enemy if you changed targets due to Dispel Magic
	if ( retarget_enemy or retarget_friend ) then
		TargetLastTarget();
	end

end

-- Make a tooltip for the BeneCastButton based on the options set
function BeneCastButton_OnEnter()

	if ( this.spelltype == 'assist' ) then
		GameTooltip:SetOwner(this, 'ANCHOR_BOTTOMRIGHT');
		GameTooltip:SetText(BENECAST_STRINGS.TEXT_ASSIST);
		return;
	end
	local spell = BeneCast_ChooseSpell(this.spelltype, 'player', true, nil);
	if ( BeneCastConfig.ShowTooltips ) then
		GameTooltip:SetOwner(this, 'ANCHOR_BOTTOMRIGHT');
		GameTooltip:SetSpell(spell.id, SpellBookFrame.bookType);
	end
	if ( BeneCastConfig.ShowTooltips and BeneCastConfig.ShowTooltipsName ) then
		GameTooltip:SetOwner(this, 'ANCHOR_BOTTOMRIGHT');
		GameTooltip:SetText(spell.name);
	end

end

-- Update the cooldown for a button on events
function BeneCastButton_UpdateCooldown()

	-- If the unit for this button does not exist return
	if ( not UnitExists(_targets[this:GetParent():GetID()]) ) then
		this:GetParent():Hide();
		return;
	end
	
	-- No need to update for a hidden button
	if ( not this:IsVisible() ) then
		return;
	end
	
	local spell;
	-- Figure out this button's spelltype
	local spelltype = this.spelltype;
	-- Find an associated spell based on the spelltype
	spell = BeneCast_ChooseSpell(spelltype, 'player', 'true', nil);
	-- If there is no spell associated with the spelltype return
	if ( not spell ) then
		return;
	end
	
	-- Cooldown stuff that I'm not quite sure how exactly it all works
	local cooldown = getglobal(this:GetName()..'Cooldown');
	local iconTexture = getglobal(this:GetName()..'Icon');
	local start, duration, enable = GetSpellCooldown(spell.id, SpellBookFrame.bookType);
	CooldownFrame_SetTimer(cooldown, start, duration, enable);
	if ( enable == 1 ) then
		iconTexture:SetVertexColor(1.0, 1.0, 1.0);
	else
		iconTexture:SetVertexColor(0.4, 0.4, 0.4);
	end

end

-- Register events for BeneCastButtons
function BeneCastButton_OnLoad()

	this:RegisterEvent('SPELL_UPDATE_COOLDOWN');
	this.flashing = nil;

end

-- *****************************************************************************
-- BeneCast Button Drawing/Updating Functions
-- *****************************************************************************

-- Function used to show a BeneCast button
-- Parameters spell_type, and the first free button available
-- 'Free' buttons are buttons that are not being used
-- Returns true if a button was created
local function BeneCast_ShowButton(type, freebutton, id)

	local made_button = nil;
	-- Figure out who this button belongs too
	local member = _targets[id];
	local class = UnitClass(member);
	if not class then
		class = BENECAST_STRINGS.CLASS_WARRIOR;
	end
	local classnum = _classes[class]
	-- If the player has the spell, wants it shown, and freebutton is <= BeneCastConfig.ButtonNumber then continue
	if ( _spell_data[type] and ( ( member ~= 'player' and BeneCastConfig[classnum][type] ) or ( member == 'player' and BeneCastConfig[1][type] ) ) and freebutton <= BeneCastConfig.ButtonNumber ) then
		local button = getglobal('BeneCastPanel' .. id .. 'Button' .. freebutton);
		-- Set made_button to true
		made_button = true;
		-- Set up the texture for the button
		local icon = getglobal('BeneCastPanel' .. id .. 'Button' .. freebutton .. 'Icon');
		icon:SetTexture(_spell_data[type].texture);
		button.spelltype = type;
		button:Show();
	end

	return made_button;
	
end

-- Function to hide temporary buttons if they're not to be shown
local function BeneCast_HideButton(buttonnum, id)

	local button = getglobal('BeneCastPanel' .. id .. 'Button' .. buttonnum);
	if ( button:IsShown() ) then
		button:Hide();
	end

end

-- Function used to make a button flash if it should be flashing
-- If so, make sure that it is set to be flashing
local function BeneCast_ButtonFlash(spelltype, member, firstfreebutton, id)

	-- Check to see if this member has the spelltype in effect
	-- Iterator for buffs
	local i = 1;

	local haseffect = nil;
	-- Cycle through the Units Buffs to find if the effect is on
	while UnitBuff(member, i) do
		-- Set BeneCast_Tooltip to the buff at iterator i be sure to clear the relevant text fields first
		BeneCast_TooltipTextLeft1:SetText('');
		BeneCast_Tooltip:SetUnitBuff(member, i);
		-- If the Tooltip has text on the Left1 it may be a buff player can cast
		if ( BeneCast_TooltipTextLeft1:GetText() ) then
			if ( _spell_types[_class][BeneCast_TooltipTextLeft1:GetText()] == spelltype ) then
				haseffect = true;
				do break end
			end
		end
		i = i + 1;
	end

	if ( haseffect ) then
		getglobal('BeneCastPanel' .. id .. 'Button' .. firstfreebutton).flashing = true;
	else
		getglobal('BeneCastPanel' .. id .. 'Button' .. firstfreebutton).flashing = nil;
	end

end

-- Function used to update buttons for all possible targets
function BeneCast_UpdateButtons(id)

	-- Sometimes the id is nil
	if not id then
		return
	end
	
	-- Figure out who this frame belongs too
	local member = _targets[id];
	
	-- If the player is dead then hide this panel and return
	if ( UnitIsDeadOrGhost('player') ) then
		getglobal('BeneCastPanel' .. id):Hide();
		return;
	end

	-- Hide this BeneCastPanel frame and return if this member is not valid
	if ( not UnitExists(member) or not UnitIsFriend('player', member) or UnitIsGhost(member) or not UnitIsConnected(member) ) then
		getglobal('BeneCastPanel' .. id):Hide();
		return;
	end
	
	-- Hide this BeneCastPanel frame and return if this member is a raidid that is not being shown
	if ( BeneCast_UnitIsRaidID(member) and ( not BENECAST_RAID_LIST[member] or not BeneCastConfig.ShowRaid ) ) then
		getglobal('BeneCastPanel' .. id):Hide();
		return;
	end
	
	-- Hide this BeneCastPanel frame and return if this member is a pet and ShowPet is not enabled
	if ( BeneCast_UnitIsPetID(member) and not BeneCastConfig.ShowPet ) then
		getglobal('BeneCastPanel' .. id):Hide();
		return;
	end
	
	-- Hide this BeneCastPanel frame and return if this member is a raid pet
	-- Until I figure out what to do with raid pets I'll just ignore them
	if ( BeneCast_UnitIsRaidPetID(member) ) then
		getglobal('BeneCastPanel' .. id):Hide();
		return;
	end
	
	-- If the player can ressurect this member then show the res button and return
	if ( UnitIsDead(member) and not UnitBuff(member, 1) ) then
		if ( _spell_data['res'] and BeneCast_ShowButton('res', 1, id) ) then
			getglobal('BeneCastPanel' .. id):Show();
			-- Resize the BeneCastPanelFrame since there's only 1 button
			local frame = getglobal('BeneCastPanel' .. id);
			local buttonsize = BeneCastConfig.ButtonSize;
			frame:SetWidth(buttonsize * (1) + 11);
			frame:SetHeight(buttonsize + 11);
			-- Hide the rest of the buttons in the panel
			for x = 2, BENECAST_MAX_NUMBER_OF_BUTTONS do
				BeneCast_HideButton(x, id);
			end
		else
			getglobal('BeneCastPanel' .. id):Hide();
		end
		return;
	end
	
	-- Hide this BeneCastPanelFrame and return if this member is a raid member and is not enabled
	if ( BeneCast_UnitIsRaidID(member) and not BENECAST_RAID_LIST[member] ) then
		getglobal('BeneCastPanel' .. id):Hide();
		return;
	end
			
	-- If player is a Druid and is shapeshifted hide all buttons
	if ( _class == BENECAST_STRINGS.CLASS_DRUID ) then
		-- Check all shapeshift forms to see if they're active
		local shapeshiftnum = GetNumShapeshiftForms();
		local icon, name, active;
		local is_shapeshifted = nil;
		for i = 1,shapeshiftnum do
			icon, name, active = GetShapeshiftFormInfo(i);
			if ( active ) then
				is_shapeshifted = true;
				do break end
			end
		end
		-- If shapeshifted hide the buttons
		if ( is_shapeshifted ) then
			getglobal('BeneCastPanel' .. id):Hide();
			return;
		end
	end
	
	-- Otherwise show the buttons. Show the BeneCastPanel frame first
	getglobal('BeneCastPanel' .. id):Show();
	
	-- firstfreebutton determines the first button that is not used
	local firstfreebutton = 1;
	
	-- Hide all buttons before redrawing them
	for x = firstfreebutton, BENECAST_MAX_NUMBER_OF_BUTTONS do
		BeneCast_HideButton(x, id);
	end
	
	-- If there is no class, set it to warrior
	local class = UnitClass(member);
	if not class then
		class = BENECAST_STRINGS.CLASS_WARRIOR;
	end
	
	-- Show the assist button if it is wanted
	if ( BeneCastConfig[_classes[class]]['assist'] and firstfreebutton <= BeneCastConfig.ButtonNumber and not UnitIsUnit(member,'player') and UnitIsPlayer(member) ) then
		local button = getglobal('BeneCastPanel' .. id .. 'Button' .. firstfreebutton);
		-- Set up the texture for the button
		local icon = getglobal('BeneCastPanel' .. id .. 'Button' .. firstfreebutton .. 'Icon');
		icon:SetTexture('Interface\\Icons\\Ability_Marksmanship');
		button.spelltype = 'assist';
		button:Show();
		firstfreebutton = firstfreebutton +1;
	end
	
	-- Show Heal buttons as persistant buttons
	if ( BeneCast_ShowButton('efficient', firstfreebutton, id) ) then
		firstfreebutton = firstfreebutton +1;
	end
	if BeneCast_ShowButton('emergency', firstfreebutton, id) then
		-- If you're a druid, check to see if this spell should be flashing
		if ( _class == BENECAST_STRINGS.CLASS_DRUID ) then
			BeneCast_ButtonFlash('emergency', member, firstfreebutton, id)
		end
		firstfreebutton = firstfreebutton +1;
	end
	if ( BeneCast_ShowButton('instant', firstfreebutton, id) ) then
		-- If you're a druid or priest, check to see if this spell should be flashing
		if _class == BENECAST_STRINGS.CLASS_DRUID or _class == BENECAST_STRINGS.CLASS_PRIEST then
			BeneCast_ButtonFlash('instant', member, firstfreebutton, id)
		end
		firstfreebutton = firstfreebutton +1;
	end
	-- Paladin's Lay on Hands spell
	if ( BeneCast_ShowButton('loh', firstfreebutton, id) ) then
		firstfreebutton = firstfreebutton +1;
	end

	-- Group heals are handled differently than other heals
	-- Only on the Shaman Group heal does targeting matter
	-- Show up only on player unless class is Shaman
	if ( _class == BENECAST_STRINGS.CLASS_SHAMAN or member == 'player' ) then
		if ( BeneCast_ShowButton('group', firstfreebutton, id) ) then
			firstfreebutton = firstfreebutton + 1;
		end
	end
	
	-- Check to see if the member can be cured of a harmful effect
	-- Do this before showing buffs so Weakened Soul is detected before drawing buffs
	-- Iterator for debuffs
	local i = 1;
	
	-- Variables to determine if cures should be shown
	-- This is done in case there are multiples of one kind of effect
	local poisonshow = nil;
	local diseaseshow = nil;
	local magicshow = nil;
	local curseshow = nil;
	
	local weakenedsoul = nil;
	
	-- Cycle through the Units Debuffs to find removable effects
	while ( UnitDebuff(member, i) ) do

		-- Set BeneCast_Tooltip to the debuff at iterator i, be sure to clear the relevant text first
		BeneCast_TooltipTextRight1:SetText('');
		BeneCast_TooltipTextLeft1:SetText('');
		BeneCast_Tooltip:SetUnitDebuff(member, i);
		-- Check to see if this debuff is Weakened Soul and save it for when you get to the buffs
		if ( BeneCast_TooltipTextLeft1:GetText() == BENECAST_STRINGS.AILMENT_WEAKENED_SOUL ) then
			weakenedsoul = true;
		end
		-- If the Tooltip has text on the Right1 it may be removable
		if ( BeneCast_TooltipTextRight1:GetText() ) then
			local effect = BeneCast_TooltipTextRight1:GetText();
			if effect == BENECAST_STRINGS.AILMENT_POISON then
				poisonshow = true;
			elseif ( effect == BENECAST_STRINGS.AILMENT_DISEASE ) then
				-- Paladins can cure diseases but don't have a spell with the 'disease' spelltype
				-- The Paladin spell that cures poisons also cures diseases
				if ( _class == BENECAST_STRINGS.CLASS_PALADIN ) then
					poisonshow = true;
				else
					diseaseshow = true;
				end
			elseif ( effect == BENECAST_STRINGS.AILMENT_MAGIC ) then
				-- Paladins can cure magic with Cleanse which is of the 'poison' spelltype
				-- If the Paladin has Cleanse have it show Cleanse
				if ( _class == BENECAST_STRINGS.CLASS_PALADIN and _spell_data['poison']['spells'][2] ) then
					poisonshow = true;
				else
					magicshow = true;
				end
			elseif ( effect == BENECAST_STRINGS.AILMENT_CURSE ) then
				curseshow = true;
			end
		end
		i = i + 1;
	end
	
	-- Show Buff buttons as temporary buttons
	
	-- See if the groupbuff is on the intended target
	local showgroupbuff = nil;
	
	-- See if a mage's ward is on the player
	local wardshown = nil;
	
	-- See if a paladin's greater blessing is on the target
	-- Thanks to Hiroko of Argent Dawn(EU)
	local gmightshown = nil;
	local gwisdomshown = nil;
	local gkingsshown = nil;
	local gsalvshown = nil;
	local glightshown = nil;
	local gsanctshown = nil;
		
	-- Check to see if this member does not have castable buffs
	
	-- Reset the iterator for buffs
	i = 1;
	
	-- Clear all entries in _buffs_in_effect
	for j in _buffs_in_effect do
		_buffs_in_effect[j].fade = false;
		_buffs_in_effect[j].ineffect = false;
	end
	
	-- Set clearcasting state to nil
	_clearcasting = nil;
	
	-- Cycle through the Units Buffs to find if the effect is on the unit
	while ( UnitBuff(member, i) ) do
				
		-- Set BeneCast_Tooltip to the buff at iterator i, be sure to clear the relevant text first
		BeneCast_TooltipTextLeft1:SetText('');
		BeneCast_Tooltip:SetUnitBuff(member, i);
		-- If the Tooltip has text on the Left1 it may be a buff player can cast
		if ( BeneCast_TooltipTextLeft1:GetText() ) then
			-- If the buff is Clearcasting then set clearcasting state to true 
			if ( BeneCast_TooltipTextLeft1:GetText() == BENECAST_STRINGS.BUFF_CLEARCASTING ) then
				_clearcasting = true;
			end
			local effect = _spell_types[_class][BeneCast_TooltipTextLeft1:GetText()];
			-- If the effect is a groupbuff set showgroupbuff to true
			if ( ( _class == BENECAST_STRINGS.CLASS_PRIEST or _class == BENECAST_STRINGS.CLASS_DRUID or _class == BENECAST_STRINGS.CLASS_MAGE ) and effect == 'groupbuff' ) then
				showgroupbuff = true;
			-- If the effect is a mage ward set wardshown to true
			elseif ( _class == BENECAST_STRINGS.CLASS_MAGE and effect == 'selfbuff12' or effect == 'selfbuff13' ) then
				wardshown = true;
			-- If the effect is a paladin greater blessing
			elseif ( _class == BENECAST_STRINGS.CLASS_PALADIN and effect == 'buffparty5') then
				gmightshown = true;
			elseif ( _class == BENECAST_STRINGS.CLASS_PALADIN and effect == 'buffparty8') then
				gwisdomshown = true;
			elseif ( _class == BENECAST_STRINGS.CLASS_PALADIN and effect == 'buffparty9') then
				gkingsshown = true;
			elseif ( _class == BENECAST_STRINGS.CLASS_PALADIN and effect == 'buffparty7') then
				gsalvshown = true;
			elseif ( _class == BENECAST_STRINGS.CLASS_PALADIN and effect == 'buffparty6') then
				glightshown = true;
			elseif ( _class == BENECAST_STRINGS.CLASS_PALADIN and effect == 'buffparty10') then
				gsanctshown = true;
            end

			-- Check the current buff at i against every buff in _buffs_in_effect
			for j in _buffs_in_effect do
				-- Check to see if the effect name is the same as the buff type
				if ( effect == _buffs_in_effect[j].name ) then
					-- Set ineffect or fade for the _buffs_in_effect entry to true depending on ShowAllBuffs
					if ( BeneCastConfig.ShowAllBuffs ) then
						-- Set fade for the type to true
						_buffs_in_effect[j].fade = true;
					else
						-- Set ineffect for the type to true
						_buffs_in_effect[j].ineffect = true;
					end
				end
			end
		end
		i = i + 1;
		
	end
	
	-- Make _buff_shown for PW:F, AI, or MotW on the intended target true if they have the groupbuff on them
	-- Set fade to true if ShowAllBuffs, else set shown to true
	if ( _class == BENECAST_STRINGS.CLASS_PRIEST and showgroupbuff ) then
		for j in _buffs_in_effect do
			if ( _buffs_in_effect[j].name == 'buff2' ) then
				if ( BeneCastConfig.ShowAllBuffs ) then
					_buffs_in_effect[j].fade = true;
				else
					_buffs_in_effect[j].ineffect = true;
				end
			end
		end
	elseif ( _class == BENECAST_STRINGS.CLASS_DRUID and showgroupbuff ) then
		for j in _buffs_in_effect do
			if ( _buffs_in_effect[j].name == 'buff1' ) then
				if ( BeneCastConfig.ShowAllBuffs ) then
					_buffs_in_effect[j].fade = true;
				else
					_buffs_in_effect[j].ineffect = true;
				end
			end
		end
	elseif ( _class == BENECAST_STRINGS.CLASS_MAGE and showgroupbuff ) then
		for j in _buffs_in_effect do
			if ( _buffs_in_effect[j].name == 'buff1' ) then
				if ( BeneCastConfig.ShowAllBuffs ) then
					_buffs_in_effect[j].fade = true;
				else
					_buffs_in_effect[j].ineffect = true;
				end
			end
		end
	elseif ( _class == BENECAST_STRINGS.CLASS_MAGE and wardshown ) then
		for j in _buffs_in_effect do
			if ( _buffs_in_effect[j].name == 'selfbuff12' or _buffs_in_effect[j].name == 'selfbuff13' ) then
				if ( BeneCastConfig.ShowAllBuffs ) then
					_buffs_in_effect[j].fade = true;
				else
					_buffs_in_effect[j].ineffect = true;
				end
			end
		end
	elseif ( _class == BENECAST_STRINGS.CLASS_PALADIN and gmightshown ) then
		for j in _buffs_in_effect do
			if ( _buffs_in_effect[j].name == 'buff1' ) then
				if ( BeneCastConfig.ShowAllBuffs ) then
					_buffs_in_effect[j].fade = true;
				else
					_buffs_in_effect[j].ineffect = true;
				end
			end
		end
	elseif ( _class == BENECAST_STRINGS.CLASS_PALADIN and glightshown ) then
		for j in _buffs_in_effect do
			if ( _buffs_in_effect[j].name == 'buff3' ) then
				if ( BeneCastConfig.ShowAllBuffs ) then
					_buffs_in_effect[j].fade = true;
				else
					_buffs_in_effect[j].ineffect = true;
				end
			end
		end
	elseif ( _class == BENECAST_STRINGS.CLASS_PALADIN and gsalvshown ) then
		for j in _buffs_in_effect do
			if ( _buffs_in_effect[j].name == 'buffparty2' ) then
				if ( BeneCastConfig.ShowAllBuffs ) then
					_buffs_in_effect[j].fade = true;
				else
					_buffs_in_effect[j].ineffect = true;
				end
			end
		end
	elseif ( _class == BENECAST_STRINGS.CLASS_PALADIN and gwisdomshown ) then
		for j in _buffs_in_effect do
			if ( _buffs_in_effect[j].name == 'buff5' ) then
				if ( BeneCastConfig.ShowAllBuffs ) then
					_buffs_in_effect[j].fade = true;
				else
					_buffs_in_effect[j].ineffect = true;
				end
			end
		end
	elseif ( _class == BENECAST_STRINGS.CLASS_PALADIN and gkingsshown ) then
		for j in _buffs_in_effect do
			if ( _buffs_in_effect[j].name == 'buff6' ) then
				if ( BeneCastConfig.ShowAllBuffs ) then
					_buffs_in_effect[j].fade = true;
				else
					_buffs_in_effect[j].ineffect = true;
				end
			end
		end
	elseif ( _class == BENECAST_STRINGS.CLASS_PALADIN and gsanctshown ) then
		for j in _buffs_in_effect do
			if ( _buffs_in_effect[j].name == 'buff4' ) then
				if ( BeneCastConfig.ShowAllBuffs ) then
					_buffs_in_effect[j].fade = true;
				else
					_buffs_in_effect[j].ineffect = true;
				end
			end
		end
	end
	
	-- Show buff buttons of buffs not in effect if possible
	for j in _buffs_in_effect do
		-- Some buttons should not show on others
		-- Selfbuffs on others and buffparty on those outside of the party/raid
		-- Groupbuffs such as Prayer of Fortitude should only be shown on the target/player
		if ( string.find(_buffs_in_effect[j].name, 'groupbuff') and ( member ~= target or member ~= 'player' ) ) then
		elseif ( string.find(_buffs_in_effect[j].name, 'selfbuff') and member ~= 'player' ) then
		elseif ( string.find(_buffs_in_effect[j].name, 'buffparty') and ( not UnitInParty(member) and not BeneCast_UnitInRaid(member) and not BeneCast_UnitIsPartyPet(member) and not BeneCast_UnitIsRaidPet(member) ) ) then
		else
			-- If the buff isn't present then show the button for that buff
			if ( not _buffs_in_effect[j].ineffect and BeneCast_ShowButton(_buffs_in_effect[j].name, firstfreebutton, id) ) then
				-- Make this button hide if it is for PW:S and the effect Weakened Soul is on the target
				if ( _class == BENECAST_STRINGS.CLASS_PRIEST and _buffs_in_effect[j].name == 'buffparty1' and weakenedsoul and not BeneCastConfig.ShowAllBuffs ) then
					getglobal('BeneCastPanel' .. id .. 'Button' .. firstfreebutton):Hide()
					firstfreebutton = firstfreebutton - 1;
				end
				-- Make the button shaded gray if the effect is on the intended target
				if ( _buffs_in_effect[j].fade ) then
					getglobal('BeneCastPanel' .. id .. 'Button' .. firstfreebutton .. 'Fade'):Show();
				else
					getglobal('BeneCastPanel' .. id .. 'Button' .. firstfreebutton .. 'Fade'):Hide();
				end
				firstfreebutton = firstfreebutton + 1;
			end
		end
	end
	
	-- Show Weapon Enchants for Shaman and Priests
	if ( member == 'player') then
		if ( _class == BENECAST_STRINGS.CLASS_SHAMAN ) then
			if ( _weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_ROCKBITER and BeneCast_ShowButton('weaponenchant1', firstfreebutton, id) ) then
				firstfreebutton = firstfreebutton+1;
			end
			if ( _weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_FLAMETONGUE and BeneCast_ShowButton('weaponenchant2', firstfreebutton, id) ) then
				firstfreebutton = firstfreebutton+1;
			end
			if ( _weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_FROSTBRAND and BeneCast_ShowButton('weaponenchant3', firstfreebutton, id) ) then
				firstfreebutton = firstfreebutton+1;
			end
			if ( _weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_WINDFURY and BeneCast_ShowButton('weaponenchant4', firstfreebutton, id) ) then
				firstfreebutton = firstfreebutton+1;
			end
		end
		if ( _class == BENECAST_STRINGS.CLASS_PRIEST ) then
			if ( _weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_FEEDBACK and BeneCast_ShowButton('weaponenchant1', firstfreebutton, id) ) then
				firstfreebutton = firstfreebutton+1;
			end
		end
	end
	
	-- Show Cure buttons as temporary buttons
		
	-- Show buttons for effects that are removable once
	if ( poisonshow and BeneCast_ShowButton('poison', firstfreebutton, id) ) then
		firstfreebutton = firstfreebutton+1;
	end
	if ( diseaseshow and BeneCast_ShowButton('disease', firstfreebutton, id) ) then
		firstfreebutton = firstfreebutton+1;
	end
	if ( magicshow and BeneCast_ShowButton('magic', firstfreebutton, id) ) then
		firstfreebutton = firstfreebutton+1;
	end
	if ( curseshow and BeneCast_ShowButton('curse', firstfreebutton, id) ) then
		firstfreebutton = firstfreebutton+1;
	end

	-- Hide all buttons that shouldn't be shown
	for x = firstfreebutton, BeneCastConfig.ButtonNumber do
		BeneCast_HideButton(x, id);
	end
	
	-- If there are no buttons to be shown, hide the frame
	if ( firstfreebutton == 1 ) then
		getglobal('BeneCastPanel' .. id):Hide();
	end
	
	-- Resize the BeneCastPanelFrame based on the number of buttons
	local frame = getglobal('BeneCastPanel' .. id);
	local buttonsize = BeneCastConfig.ButtonSize;
	frame:SetWidth(buttonsize * (firstfreebutton - 1) + 11 + (3 * (firstfreebutton - 2)));
	frame:SetHeight(buttonsize + 11);
		
end


-- *****************************************************************************
-- BeneCastPanelMenu functions
-- *****************************************************************************

function BeneCastPanelDropDown_Initialize()
	
	local info;
	-- Add Title Button
	info = {};
	if ( UnitExists(_targets[this:GetID()]) ) then
		info.text = UnitName(_targets[this:GetID()]);
	else
		info.text = BENECAST_STRINGS.TEXT_UNKNOWN
	end
	info.isTitle = 1;
	UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
	
	-- Add Lock Button
	info = {};
	info.text = BENECAST_STRINGS.TEXT_LOCKED;
	info.value = this:GetID();
	info.checked = this.locked;
	info.func = BeneCastPanel_SetLock;
	UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);

end

function BeneCastPanel_SetLock()

	local frame = getglobal('BeneCastPanel' .. this.value);
	frame.locked = not frame.locked;
	
end

function BeneCastPanelMenu_OnLoad()

	UIDropDownMenu_Initialize(this, BeneCastPanelDropDown_Initialize, 'MENU');

end

-- *****************************************************************************
-- BeneCastPanel functions
-- *****************************************************************************

function BeneCastPanel_OnLoad()

	-- set our default colors
	this:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b);
	this:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b);
	this.locked = 1;

end

function BeneCastPanel_OnMouseDown()

	if ( ( ( not this.locked ) or ( this.locked == 0 ) or BeneCastConfig.Unlock ) and ( arg1 == 'LeftButton' ) ) then
		this:StartMoving();
		this.isMoving = true;
	end
	
end

-- *****************************************************************************
-- BeneCastOptionFrameTab functions
-- *****************************************************************************

function BeneCastOptionFrameTab_OnClick()

	-- Show the correct subframe based on which tab was clicked
	-- Hide all subframes that do not have matching ID numbers, show the matching subframe
	for subframe in BeneCastOptionFrameSubframes do
		if this:GetID() == getglobal(BeneCastOptionFrameSubframes[subframe]):GetID() then
			getglobal(BeneCastOptionFrameSubframes[subframe]):Show();
		else
			getglobal(BeneCastOptionFrameSubframes[subframe]):Hide();
		end
	end
	
	-- Play a sound for player feedback
	PlaySound('igCharacterInfoTab');
	
	-- Set the glowing checked status for the correct tab
	for x = 1, 12 do
		if ( x == this:GetID() ) then
			this:Disable()
			this:SetChecked(1);
		else
			getglobal('BeneCastOptionFrameTab' .. x):Enable();
			getglobal('BeneCastOptionFrameTab' .. x):SetChecked(nil);
		end
	end
	
end

-- *****************************************************************************
-- BeneCastOptionsCheckButton functions
-- *****************************************************************************

function BeneCastOptionsCheckButton_OnClick()

	-- Set the BeneCastConfig for the option to take on the value of the checkbutton
	for index, value in BeneCastOptionFrameCheckButtons do
		if ( value.index == this:GetID() ) then
			BeneCastConfig[value.cvar] = this:GetChecked();
			-- If this is the tooltip option check to see if it's enabled
			-- If not disable the tooltip name option checkbutton
			if ( value.index == 5 ) then
				if ( this:GetChecked() ) then
					getglobal('BeneCastOptionCheckButton6'):Enable();
				else
					getglobal('BeneCastOptionCheckButton6'):Disable();
				end
			-- If this is the notification option check to see if it's enabled
			-- If not disable all sub notification options
			elseif ( value.index == 8 ) then
				if ( this:GetChecked() ) then
					getglobal('BeneCastOptionCheckButton9'):Enable();
					getglobal('BeneCastOptionCheckButton10'):Enable();
					getglobal('BeneCastOptionCheckButton11'):Enable();
					getglobal('BeneCastOptionCheckButton12'):Enable();
					getglobal('BeneCastOptionCheckButton20'):Enable();
					getglobal('BeneCastOptionCheckButton13'):Enable();
					getglobal('BeneCastOptionCheckButton14'):Enable();
					getglobal('BeneCastOptionCheckButton15'):Enable();
					getglobal('BeneCastOptionCheckButton16'):Enable();
					getglobal('BeneCastOptionCheckButton17'):Enable();
					getglobal('BeneCastOptionCheckButton18'):Enable();
					getglobal('BeneCastOptionCheckButton19'):Enable();
					getglobal('BeneCastOptionCheckButton21'):Enable();
				else
					getglobal('BeneCastOptionCheckButton9'):Disable();
					getglobal('BeneCastOptionCheckButton10'):Disable();
					getglobal('BeneCastOptionCheckButton11'):Disable();
					getglobal('BeneCastOptionCheckButton12'):Disable();
					getglobal('BeneCastOptionCheckButton20'):Disable();
					getglobal('BeneCastOptionCheckButton13'):Disable();
					getglobal('BeneCastOptionCheckButton14'):Disable();
					getglobal('BeneCastOptionCheckButton15'):Disable();
					getglobal('BeneCastOptionCheckButton16'):Disable();
					getglobal('BeneCastOptionCheckButton17'):Disable();
					getglobal('BeneCastOptionCheckButton18'):Disable();
					getglobal('BeneCastOptionCheckButton19'):Disable();
					getglobal('BeneCastOptionCheckButton21'):Disable();
				end
			-- If this is the heal notification option check to see if it's enabled
			-- If not disable all heal max notification
			elseif ( value.index == 16 ) then
				if ( this:GetChecked() ) then
					getglobal('BeneCastOptionCheckButton17'):Enable();
				else
					getglobal('BeneCastOptionCheckButton17'):Disable();
				end
			-- If this is the hide minimap button option check to see if it's enabled
			-- If so then hide the minimap button
			elseif ( value.index == 4 ) then
				if ( this:GetChecked() ) then
					getglobal('BeneCastMinimapButton'):Hide();
				else
					getglobal('BeneCastMinimapButton'):Show();
				end	
			-- If this is a notification channel option check to see if it's enabled
			-- If so disable all other notification channel options
			-- Notifcation channel options should act like radio buttons
			elseif ( value.index == 9 or value.index == 10 or value.index == 11 or value.index == 12 or value.index == 20 ) then
				if ( this:GetChecked() ) then
					-- Uncheck all other buttons
					getglobal('BeneCastOptionCheckButton9'):SetChecked(0);
					getglobal('BeneCastOptionCheckButton10'):SetChecked(0);
					getglobal('BeneCastOptionCheckButton11'):SetChecked(0);
					getglobal('BeneCastOptionCheckButton12'):SetChecked(0);
					getglobal('BeneCastOptionCheckButton20'):SetChecked(0);
					-- Check this button
					getglobal('BeneCastOptionCheckButton'..value.index):SetChecked(1)
					-- Set all the variables to the correct values
					-- Set all the channels to false
					BeneCastConfig['SelfNotify'] = false;
					BeneCastConfig['PartyNotify'] = false;
					BeneCastConfig['RaidNotify'] = false;
					BeneCastConfig['TargetNotify'] = false;
					BeneCastConfig['UserNotify'] = false;
					-- Set the correct channel to true
					BeneCastConfig[value.cvar] = 1;
				else
					-- Don't let the user uncheck a button by clicking on it
					-- It will be unchecked if another channel is set
					-- As such, the config setting must still be set to 1
					BeneCastConfig[value.cvar] = 1;
					getglobal('BeneCastOptionCheckButton'..value.index):SetChecked(1)
				end
			end
		end
	end
	-- Play a sound for player feedback
	if ( this:GetChecked() ) then
		PlaySound('igMainMenuOptionCheckBoxOff');
	else
		PlaySound('igMainMenuOptionCheckBoxOn');
	end
	
	-- Redraw everyone's buttons
	for i in _targets do
		BeneCast_UpdateButtons(i);
	end

end


-- *****************************************************************************
-- BeneCastSpellCheckButton functions
-- *****************************************************************************

function BeneCastSpellCheckButton_OnClick()

	-- Set the BeneCastConfig for the option to take on the value of the checkbutton
	if ( getglobal( this:GetName() .. 'Text' ):GetText() == BENECAST_STRINGS.TEXT_ASSIST ) then
		BeneCastConfig[this:GetParent():GetID()]['assist'] = this:GetChecked();
	else
		if ( BeneCastConfig['PlayerAsDefault'] and this:GetParent():GetID() == 1 ) then
			for i = 1, 9 do
				BeneCastConfig[i][_spell_types[_class][getglobal( this:GetName() .. 'Text' ):GetText()]] = this:GetChecked();
				getglobal(BeneCastOptionFrameSubframes[i] .. 'Button' .. this:GetID()):SetChecked(this:GetChecked());
			end
		else
			BeneCastConfig[this:GetParent():GetID()][_spell_types[_class][getglobal( this:GetName() .. 'Text' ):GetText()]] = this:GetChecked();
		end
	end
	-- Play a sound for player feedback
	if ( this:GetChecked() ) then
		PlaySound('igMainMenuOptionCheckBoxOff');
	else
		PlaySound('igMainMenuOptionCheckBoxOn');
	end
	
	-- Redraw everyone's buttons
	for i in _targets do
		BeneCast_UpdateButtons(i);
	end

end

-- *****************************************************************************
-- BeneCastRaidCheckButton functions
-- *****************************************************************************

function BeneCastRaidCheckButton_OnClick()

	-- If this raid checkbutton has a name
	if ( getglobal(this:GetName() .. 'Name'):GetText() ~= '' ) then
		-- Set the RaidID of this member to on or off
		local unitid;
		local name = getglobal(this:GetName() .. 'Name'):GetText();
		local idnum = 1;
		for i = 1, MAX_RAID_MEMBERS do
			if UnitName('raid'..i) == name then
				unitid = 'raid'..i;
				idnum = i;
				break;
			end
		end
		BENECAST_RAID_LIST[unitid] = this:GetChecked();
		
		BeneCast_AttachRaidPanels(BeneCastConfig.RaidSnapTo);
		
		-- Update the buttons associated with the raid member
		BeneCast_UpdateButtons(BeneCast_UnitID(unitid));
		
	end

end

function BeneCastRaidCheckButton_OnEnter()

	GameTooltip:SetOwner(this, 'ANCHOR_BOTTOMRIGHT');
	GameTooltip:SetText(BENECAST_STRINGS.TEXT_RAID_CHECKBUTTON_TOOLTIP, nil, nil, nil, nil, 1)

end

-- *****************************************************************************
-- BeneCastRaidGroup functions
-- *****************************************************************************

function BeneCastRaidGroup_OnClick()

	local name, rank, subgroup, level, class, fileName, zone, online, isDead;
	
	-- For all raid members/ids
	for i = 1, MAX_RAID_MEMBERS do
		name, rank, subgroup, level, class, fileName, zone, online, isDead = GetRaidRosterInfo(i);
		-- Check this raid member to if their subgroup is the same as this subgroup
		if ( subgroup	== this:GetID() ) then
			if ( name ) then
				-- Set the RaidID of this group member to on
				BENECAST_RAID_LIST['raid' .. i] = true;
				
				-- Update the buttons associated with the raid member
				BeneCast_UpdateButtons(BeneCast_UnitID('raid' .. i));
		
				-- Check all the buttons in this group
				getglobal('BeneCastRaidButton' .. ((subgroup - 1) * 5) + 1 ):SetChecked(true);
				getglobal('BeneCastRaidButton' .. ((subgroup - 1) * 5) + 2 ):SetChecked(true);
				getglobal('BeneCastRaidButton' .. ((subgroup - 1) * 5) + 3 ):SetChecked(true);
				getglobal('BeneCastRaidButton' .. ((subgroup - 1) * 5) + 4 ):SetChecked(true);
				getglobal('BeneCastRaidButton' .. ((subgroup - 1) * 5) + 5 ):SetChecked(true);
			end
		end
	end
	-- Attach the panels to the appropriate frames
	BeneCast_AttachRaidPanels(BeneCastConfig.RaidSnapTo);
	
end

function BeneCastRaidGroup_OnLoad()

	getglobal(this:GetName() .. 'Name'):SetText(BENECAST_STRINGS.TEXT_RAID_GROUP .. this:GetID());

end

function BeneCastRaidGroup_OnEnter()

	GameTooltip:SetOwner(this, 'ANCHOR_BOTTOMRIGHT');
	GameTooltip:SetText(BENECAST_STRINGS.TEXT_RAID_GROUP_TOOLTIP, nil, nil, nil, nil, 1)

end

-- *****************************************************************************
-- BeneCastOptionsSlider functions
-- *****************************************************************************

function BeneCastOptionsSlider_OnValueChanged()

	-- Set the BeneCastConfig for the option to take on the value of the checkbutton
	for index, value in BeneCastOptionFrameSliders do
		if ( value.index == this:GetID() ) then
			BeneCastConfig[value.cvar] = this:GetValue();
			getglobal('BeneCastOptionSlider'..value.index..'Value'):SetText(this:GetValue());
		end
		-- If the slider being changed is the buttonsize slider then resize the buttons
		if ( value.index == 1 ) then
			BeneCast_ResizeButtons();
		end
		-- If the slider being changed is the buttonnumber slider then redraw all buttons
		if ( value.index == 3 ) then
			for i in _targets do
				BeneCast_UpdateButtons(i);
			end
		end
	end

end


-- *****************************************************************************
-- BeneCastPanelManager functions
-- *****************************************************************************

function BeneCastPanelManager_OnLoad()

	-- Set _class to the player's class
	_class = UnitClass('player');
	
	-- Set _name to the player's name
	_name = UnitName('player');
	
	-- Register for Events
	this:RegisterEvent('ADDON_LOADED');
	this:RegisterEvent('SPELLS_CHANGED');
	this:RegisterEvent('SPELLCAST_FAILED');
	this:RegisterEvent('SPELLCAST_INTERRUPTED');
	this:RegisterEvent('SPELLCAST_STOP');
	this:RegisterEvent('SPELLCAST_START');
	this:RegisterEvent('PARTY_MEMBERS_CHANGED');
	this:RegisterEvent('PLAYER_TARGET_CHANGED');
	this:RegisterEvent('UNIT_AURA');
	this:RegisterEvent('UPDATE_SHAPESHIFT_FORMS');
	this:RegisterEvent('PLAYER_ALIVE');
	this:RegisterEvent('PLAYER_DEAD');
	this:RegisterEvent('PLAYER_UNGHOST');
	this:RegisterEvent('RAID_ROSTER_UPDATE');
	this:RegisterEvent('UNIT_HEALTH');

end

function BeneCastPanelManager_OnEvent()

	if ( event == 'ADDON_LOADED' and arg1 == 'BeneCast' ) then
		-- Reset the data if there BeneCast is severly outdated
		if ( not BeneCastConfig or not BeneCastConfig.Version or BeneCastConfig.Version < BENECAST_VERSION ) then
			BeneCast_ResetConfig();
		end
		-- Load up spell data once AddOn has been loaded
		BeneCast_ResetData();
		
	-- The following is really, really confusing
	-- If SPELLCAST_FAILED then a BeneCast spell may have failed
	elseif ( event == 'SPELLCAST_FAILED' ) then
		-- A BeneCast spell failed if a _spellstarted and _sentmessage
		if ( _spellstarted and _sentmessage ) then
			BeneCast_PartyNotifyFail('SPELLCAST_FAILED');
		end
		-- Clear out everything used for party notification just in case
		_spellstarted = nil;
		_spellstopped = nil;
		_spelltarget = nil;
		_spellbeingcast = nil;
		_spellismax = nil;
		_sentmessage = nil;
	-- If SPELLCAST_INTERRUPTED then a BeneCast spell may have been interrupted
	elseif ( event == 'SPELLCAST_INTERRUPTED' ) then
		-- A BeneCast spell was interrupted by the player if _spellstopped and _sentmessage
		if ( _spellstopped and _sentmessage ) then
			BeneCast_PartyNotifyFail('SPELLCAST_CANCELED');
		end
		-- A BeneCast spell was interrupted by something else if _spellstarted and _sentmessage
		if ( _spellstarted and _sentmessage ) then
			BeneCast_PartyNotifyFail('SPELLCAST_INTERRUPTED');
		end
		-- Clear out everything just in case
		_spellstarted = nil;
		_spellstopped = nil;
		_spelltarget = nil;
		_spellbeingcast = nil;
		_spellismax = nil;
		_sentmessage = nil;
	elseif ( event == 'SPELLCAST_STOP' ) then
		-- If _spellbeingcast and notification is on then this is an instant spell and you should do notification
		if ( _spellbeingcast and BeneCastConfig.Notification ) then
			-- _sentmessage will be set to true of BeneCast_PartyNotify sends a message
			_sentmessage = BeneCast_PartyNotify(_spelltarget, _spellbeingcast, _spellismax, 0);
		-- Otherwise set _setmessage to false
		else
			_sentmessage = nil
		end
		-- If _spellstarted, then a BeneCast spell was started and you should set _sentmessage and _spellstopped to true to indicate that the spell was interrupted by the user
		if ( _spellstarted ) then
			_sentmessage = true;
			_spellstopped = true;
			_spellstarted = nil;
		else
			-- Clear this, just in case
			_spellstopped = nil;
		end
		-- Clear these out just in case
		_spelltarget = nil;
		_spellbeingcast = nil;
		_spellismax = nil;
	elseif ( event == 'SPELLCAST_START' ) then
		-- If the spell being cast exists in _spell_types, notification is on, and there is a _spellbeingcast then do notification
		if ( _spell_types[_class][arg1] and BeneCastConfig.Notification and _spellbeingcast ) then
			-- Set _sentmessage to true of BeneCast_PartyNotify sent a message
			_sentmessage = BeneCast_PartyNotify(_spelltarget, _spellbeingcast, _spellismax, arg2);
			-- Clear out _spelltarget, _spellbeingcast, and _spellismax so it isn't used by SPELLCAST_STOP
			_spelltarget = nil;
			_spellbeingcast = nil;
			_spellismax = nil;
			-- If you sent a message then _spellstarted to indicate failures/interruptions should send notifications
			if ( _sentmessage ) then
				_spellstarted = true;
			end
		else
			-- Clear out everything just in case
			_spellstarted = nil;
			_spellstopped = nil;
			_spelltarget = nil;
			_spellbeingcast = nil;
			_spellismax = nil;
			_sentmessage = nil;
		end
	
	-- We only want spell changes after entering the world
	elseif ( event == 'SPELLS_CHANGED' ) then
		
		-- Reset most of the tables for reloading of spells and reload the spell data
		BeneCast_ResetData();
		-- Redraw all the buttons on everyone
		for i in _targets do
			BeneCast_UpdateButtons(i);
		end
	
	-- Don't bother with UNIT_AURA if the player can attack the unit
	elseif ( event == 'UNIT_AURA' and not UnitCanAttack('player', arg1) ) then
		-- Redraw the buttons for the UnitID in arg1
		BeneCast_UpdateButtons(BeneCast_UnitID(arg1));
	
	elseif ( event == 'PARTY_MEMBERS_CHANGED' ) then
		-- Redraw the buttons for all partymembers
		for i = 2,5 do
			BeneCast_UpdateButtons(i);
		end
	
	elseif ( event == 'PLAYER_TARGET_CHANGED' ) then
		-- Redraw the buttons for the target
		BeneCast_UpdateButtons(91);
		
	elseif ( event == 'RAID_ROSTER_UPDATE' ) then
		-- Redraw the buttons for the raid
		for i = 6,45 do
			BeneCast_UpdateButtons(i);
		end
		
	elseif ( event == 'PLAYER_ALIVE' ) then
		-- Redraw everyone's buttons
		for i in _targets do
			BeneCast_UpdateButtons(i);
		end
		
	elseif ( event == 'PLAYER_DEAD' ) then
		-- Redraw everyone's buttons
		for i in _targets do
			BeneCast_UpdateButtons(i);
		end
		
	elseif ( event == 'PLAYER_UNGHOST' ) then
		-- Redraw everyone's buttons
		for i in _targets do
			BeneCast_UpdateButtons(i);
		end
		
	elseif ( event == 'UNIT_HEALTH' ) then
		-- Redraw a member's panel if they're dead
		if ( UnitHealth(arg1) <= 0 and UnitIsConnected(arg1) ) then
			BeneCast_UpdateButtons(BeneCast_UnitID(arg1));
		end
		
	elseif ( event == 'UPDATE_SHAPESHIFT_FORMS' ) then
		-- Redraw everyone's buttons
		for i in _targets do
			BeneCast_UpdateButtons(i);
		end
		
	end

end

function BeneCastPanelManager_OnUpdate()

	local hasMainHandEnchant, mainHandExpiration, mainHandCharges, hasOffHandEnchant, offHandExpiration, offHandCharges;
	
	hasMainHandEnchant, mainHandExpiration, mainHandCharges, hasOffHandEnchant, offHandExpiration, offHandCharges = GetWeaponEnchantInfo();
	
	local slotId, textureName;
	
	slotId, textureName = GetInventorySlotInfo('MainHandSlot');
	
	BeneCast_TooltipTextLeft6:SetText('');
	
	if ( hasMainHandEnchant ) then
		BeneCast_Tooltip:SetInventoryItem('player', slotId);
		if ( _weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_ROCKBITER and string.find(BeneCast_TooltipTextLeft6:GetText(), BENECAST_STRINGS.WEAPONENCHANT_ROCKBITER) ) then
			_weaponenchant = BENECAST_STRINGS.WEAPONENCHANT_ROCKBITER;
			BeneCast_UpdateButtons(1);
		end
		if ( _weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_FLAMETONGUE and string.find(BeneCast_TooltipTextLeft6:GetText(), BENECAST_STRINGS.WEAPONENCHANT_FLAMETONGUE) ) then
			_weaponenchant = BENECAST_STRINGS.WEAPONENCHANT_FLAMETONGUE;
			BeneCast_UpdateButtons(1);
		end
		if ( _weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_FROSTBRAND and string.find(BeneCast_TooltipTextLeft6:GetText(), BENECAST_STRINGS.WEAPONENCHANT_FROSTBRAND) ) then
			_weaponenchant = BENECAST_STRINGS.WEAPONENCHANT_FROSTBRAND;
			BeneCast_UpdateButtons(1);
		end
		if ( _weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_WINDFURY and string.find(BeneCast_TooltipTextLeft6:GetText(), BENECAST_STRINGS.WEAPONENCHANT_WINDFURY) ) then
			_weaponenchant = BENECAST_STRINGS.WEAPONENCHANT_WINDFURY;
			BeneCast_UpdateButtons(1);
		end
		if ( _weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_FEEDBACK and string.find(BeneCast_TooltipTextLeft6:GetText(), BENECAST_STRINGS.WEAPONENCHANT_FEEDBACK) ) then
			_weaponenchant = BENECAST_STRINGS.WEAPONENCHANT_FEEDBACK;
			BeneCast_UpdateButtons(1);
		end
	else
		if ( _weaponenchant ) then
			_weaponenchant = nil;
			BeneCast_UpdateButtons(1);
		end
		_weaponenchant = nil;
	end

end

-- *****************************************************************************
-- BeneCastResetButton functions
-- *****************************************************************************

function BeneCast_ResetTooltip()

	GameTooltip:SetOwner(this, 'ANCHOR_RIGHT');
	GameTooltip:SetText(BENECAST_STRINGS['TEXT_RESET_BUTTON_TOOLTIP'], nil, nil, nil, nil, 1);

end

function BeneCast_ResetConfig()

	-- Reset the BeneCastConfig table
	BeneCastConfig = {};
	
	BeneCastConfig[1] = {};
	BeneCastConfig[2] = {};
	BeneCastConfig[3] = {};
	BeneCastConfig[4] = {};
	BeneCastConfig[5] = {};
	BeneCastConfig[6] = {};
	BeneCastConfig[7] = {};
	BeneCastConfig[8] = {};
	BeneCastConfig[9] = {};

	for i = 1,9 do
		for type in _spell_level[_class] do
			BeneCastConfig[i][type] = false;
		end
	end
	
	BeneCastConfig['DmgBasedHeal'] = true;
	BeneCastConfig['Overheal'] = false;
	BeneCastConfig['ShowAllBuffs'] = false;
	BeneCastConfig['HideMinimap'] = false;
	BeneCastConfig['ShowTooltips'] = true;
	BeneCastConfig['ShowTooltipsName'] = false;
	BeneCastConfig['Unlock'] = false;
	BeneCastConfig['Notification'] = false;
	BeneCastConfig['SelfNotify'] = true;
	BeneCastConfig['PartyNotify'] = false;
	BeneCastConfig['RaidNotify'] = false;
	BeneCastConfig['TargetNotify'] = false;
	BeneCastConfig['UserNotify'] = false;
	BeneCastConfig['SelfCasts'] = true;
	BeneCastConfig['NotifyRank'] = true;
	BeneCastConfig['NotifyTime'] = true;
	BeneCastConfig['NotifyHeal'] = true;
	BeneCastConfig['NotifyMaxHeal'] = false;
	BeneCastConfig['NotifyCure'] = true;
	BeneCastConfig['NotifyBuff'] = true;
	BeneCastConfig['NotifyRes'] = true;
	BeneCastConfig['PlayerAsDefault'] = true;
	BeneCastConfig['ShowPets'] = false;
	BeneCastConfig['ShowRaid'] = false;
	
	BeneCastConfig['ButtonSize'] = 36;
	BeneCastConfig['OverhealSlider'] = 1;
	BeneCastConfig['ButtonNumber'] = 5;
	
	BeneCastConfig['UserChannel'] = nil;
	
	BeneCastConfig['SnapTo'] = 'STANDARD';
	BeneCastConfig['SnapTo'] = 'DISCORD';
	BeneCastConfig['RaidSnapTo'] = 'STANDARD';
	
	BeneCastConfig['Version'] = BENECAST_VERSION;
	
	BeneCast_ResetData();
	BeneCast_LoadOptions();
	-- Place the panels in the correct place
	BeneCast_SnapPanels('BeneCast');
	-- Reset the SnapTo selection
	UIDropDownMenu_SetSelectedValue(BeneCastSnapToMenu, BeneCastConfig.SnapTo);
	-- Redraw everyone's panels
	for i in _targets do
		BeneCast_UpdateButtons(i);
	end

end

-- *****************************************************************************
-- BeneCastUserChannel functions
-- *****************************************************************************

function BeneCastUserChannel_OnTextChange()

	BeneCastConfig['UserChannel'] = this:GetText();

end

-- *****************************************************************************
-- BeneCastSnapToMenu functions
-- *****************************************************************************

function BeneCastSnapToMenu_Initialize()
	
	local info;
	-- Add SnapTo options
	for snapto in BeneCast_SnapTo do
		if ( IsAddOnLoaded(BeneCast_SnapTo[snapto].AddOn) ) then
			info = {};
			if ( BeneCast_SnapTo[snapto].AddOn == 'BeneCast' ) then
				info.text = BENECAST_STRINGS.TEXT_STANDARD;
			elseif ( BeneCast_SnapTo[snapto].AddOn == 'Perl_Target' ) then
				info.text = BENECAST_STRINGS.TEXT_PERLCLASSIC;
			else
				info.text = BeneCast_SnapTo[snapto].AddOn;
			end
			info.value = snapto;
			if ( BeneCastConfig.SnapTo == snapto ) then
				info.checked = true;
			else
				info.checked = false;
			end
			info.func = BeneCastSnapToMenu_SetSnap;
			UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
		end
	end

end

function BeneCastSnapToMenu_SetSnap()

	BeneCastConfig.SnapTo = this.value;
	BeneCast_SnapPanels(BeneCast_SnapTo[this.value].AddOn);
	UIDropDownMenu_SetSelectedValue(BeneCastSnapToMenu, this.value);
	for i in _targets do
		BeneCast_UpdateButtons(i);
	end
	
end

function BeneCastSnapToMenu_Load()

	UIDropDownMenu_Initialize(this, BeneCastSnapToMenu_Initialize);
	UIDropDownMenu_SetSelectedValue(this, BeneCastConfig.SnapTo);
	getglobal('BeneCastSnapToMenuLabel'):SetText(BENECAST_STRINGS.TEXT_SNAP_TO);
	BeneCastSnapToMenu.tooltip = BENECAST_STRINGS.TEXT_SNAP_TO_TOOLTIP;
	UIDropDownMenu_SetWidth(90, BeneCastSnapToMenu);

end

-- *****************************************************************************
-- BeneCastRaidSnapToMenu functions
-- *****************************************************************************

function BeneCastRaidSnapToMenu_Initialize()
	
	local info;
	-- Add SnapTo options
	for snapto in BeneCast_RaidSnapTo do
		if ( IsAddOnLoaded(BeneCast_RaidSnapTo[snapto].AddOn) ) then
			info = {};
			if ( BeneCast_RaidSnapTo[snapto].AddOn == 'BeneCast' ) then
				info.text = BENECAST_STRINGS.TEXT_STANDARD;
			else
				info.text = BeneCast_RaidSnapTo[snapto].AddOn;
			end
			info.value = snapto;
			if ( BeneCastConfig.RaidSnapTo == snapto ) then
				info.checked = true;
			else
				info.checked = false;
			end
			info.func = BeneCastRaidSnapToMenu_SetSnap;
			UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
		end
	end

end

function BeneCastRaidSnapToMenu_SetSnap()

	BeneCastConfig.RaidSnapTo = this.value;
	BeneCast_AttachRaidPanels(BeneCast_RaidSnapTo[this.value].AddOn);
	UIDropDownMenu_SetSelectedValue(BeneCastRaidSnapToMenu, this.value);
	BeneCast_UpdateRaid();
	
end

function BeneCastRaidSnapToMenu_Load()

	UIDropDownMenu_Initialize(this, BeneCastRaidSnapToMenu_Initialize);
	UIDropDownMenu_SetSelectedValue(this, BeneCastConfig.RaidSnapTo);
	getglobal('BeneCastRaidSnapToMenuLabel'):SetText(BENECAST_STRINGS.TEXT_RAID_SNAP_TO);
	BeneCastRaidSnapToMenu.tooltip = BENECAST_STRINGS.TEXT_RAID_SNAP_TO_TOOLTIP;

end

-- *****************************************************************************
-- BeneCastOptionFrame functions
-- *****************************************************************************

function BeneCastOptionFrame_OnLoad()
	
	this:RegisterEvent('ADDON_LOADED');
	this:RegisterEvent('RAID_ROSTER_UPDATE');
	
	-- Binding labels
	BINDING_HEADER_BENECAST = 'BeneCast';
	BINDING_NAME_BENECAST_OPTIONS = BENECAST_STRINGS.TEXT_BENECAST_OPTIONS;
	BINDING_NAME_BENECAST_TOGGLE_BUFFS = BENECAST_STRINGS.TEXT_BENECAST_TOGGLE_BUFFS;
	BINDING_NAME_BENECAST_BUTTON1 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON1;
	BINDING_NAME_BENECAST_BUTTON2 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON2;
	BINDING_NAME_BENECAST_BUTTON3 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON3;
	BINDING_NAME_BENECAST_BUTTON4 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON4;
	BINDING_NAME_BENECAST_BUTTON5 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON5;
	BINDING_NAME_BENECAST_BUTTON6 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON6;
	BINDING_NAME_BENECAST_BUTTON7 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON7;
	BINDING_NAME_BENECAST_BUTTON8 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON8;
	BINDING_NAME_BENECAST_BUTTON9 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON9;
	BINDING_NAME_BENECAST_BUTTON10 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON10;
	SLASH_BENECAST1 = '/benecast';
	SLASH_BENECAST2 = '/bc';
	
	-- Setup the BeneCast Slash Command Handler
	SlashCmdList['BENECAST'] = function(msg)
		BeneCast_SlashCommandHandler(msg);
	end
	
	-- Set up the title in the option frame
	getglobal('BeneCastTitleText'):SetText(BENECAST_STRINGS.TEXT_TITLE);

	-- Check Tab 1 and show the Player spell selection subframe
	getglobal('BeneCastPlayerFrame'):Show();
	getglobal('BeneCastOptionFrameTab1'):SetChecked(1);
	getglobal('BeneCastOptionFrameTab1'):Disable();
	
	-- Setup Tab Text and Graphics
	getglobal('BeneCastOptionFrameTab1NormalTexture'):SetTexture('Interface\\Icons\\Spell_Nature_WispSplode');
	getglobal('BeneCastOptionFrameTab1').tooltipText = UnitName('player');
	
	getglobal('BeneCastOptionFrameTab2NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab2NormalTexture'):SetTexCoord(0.7421875, 0.98828125, 0, 0.25);
	getglobal('BeneCastOptionFrameTab2').tooltipText = BENECAST_STRINGS.CLASS_DRUID;
	
	getglobal('BeneCastOptionFrameTab3NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab3NormalTexture'):SetTexCoord(0, 0.25, 0.25, 0.5);
	getglobal('BeneCastOptionFrameTab3').tooltipText = BENECAST_STRINGS.CLASS_HUNTER;
	
	getglobal('BeneCastOptionFrameTab4NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab4NormalTexture'):SetTexCoord(0.25, 0.49609375, 0, 0.25);
	getglobal('BeneCastOptionFrameTab4').tooltipText = BENECAST_STRINGS.CLASS_MAGE;
	
	getglobal('BeneCastOptionFrameTab5NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab5NormalTexture'):SetTexCoord(0.49609375, 0.7421875, 0.25, 0.5);
	getglobal('BeneCastOptionFrameTab5').tooltipText = BENECAST_STRINGS.CLASS_PRIEST;
	
	getglobal('BeneCastOptionFrameTab6NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab6NormalTexture'):SetTexCoord(0.49609375, 0.7421875, 0, 0.25);
	getglobal('BeneCastOptionFrameTab6').tooltipText = BENECAST_STRINGS.CLASS_ROGUE;
	
	getglobal('BeneCastOptionFrameTab7NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab7NormalTexture'):SetTexCoord(0.7421875, 0.98828125, 0.25, 0.5);
	getglobal('BeneCastOptionFrameTab7').tooltipText = BENECAST_STRINGS.CLASS_WARLOCK;
	
	getglobal('BeneCastOptionFrameTab8NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab8NormalTexture'):SetTexCoord(0, 0.25, 0, 0.25);
	getglobal('BeneCastOptionFrameTab8').tooltipText = BENECAST_STRINGS.CLASS_WARRIOR;
	
	local race, raceEn = UnitRace('player');
	
	if ( raceEn == 'Human' or raceEn == 'NightElf' or raceEn == 'Gnome' or raceEn == 'Dwarf' ) then
		getglobal('BeneCastOptionFrameTab9NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
		getglobal('BeneCastOptionFrameTab9NormalTexture'):SetTexCoord(0.0, 0.25, 0.5, 0.75);
		getglobal('BeneCastOptionFrameTab9').tooltipText = BENECAST_STRINGS.CLASS_PALADIN;
		getglobal('BeneCastOptionFrameTab12NormalTexture'):SetTexture('Interface\\Icons\\INV_Banner_02');
	else
		getglobal('BeneCastOptionFrameTab9NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
		getglobal('BeneCastOptionFrameTab9NormalTexture'):SetTexCoord(0.25, 0.49609375, 0.25, 0.5);
		getglobal('BeneCastOptionFrameTab9').tooltipText = BENECAST_STRINGS.CLASS_SHAMAN;
		getglobal('BeneCastOptionFrameTab12NormalTexture'):SetTexture('Interface\\Icons\\INV_Banner_03');
	end
	
	getglobal('BeneCastOptionFrameTab10NormalTexture'):SetTexture('Interface\\Icons\\INV_Misc_Wrench_01');
	getglobal('BeneCastOptionFrameTab10').tooltipText = BENECAST_STRINGS.TEXT_SETUP;
	
	getglobal('BeneCastOptionFrameTab11NormalTexture'):SetTexture('Interface\\Icons\\INV_Letter_08');
	getglobal('BeneCastOptionFrameTab11').tooltipText = BENECAST_STRINGS.TEXT_NOTIFICATION;
	
	getglobal('BeneCastOptionFrameTab12NormalTexture'):SetTexture('Interface\\Icons\\INV_Misc_Head_Dragon_01');
	getglobal('BeneCastOptionFrameTab12').tooltipText = BENECAST_STRINGS.TEXT_RAID;

end

function BeneCastOptionFrame_OnEvent()

	if ( event == 'ADDON_LOADED' and arg1 == 'BeneCast' ) then
		BeneCast_LoadOptions();
		BeneCast_ResizeButtons();
		BeneCast_ClearRaid();
		
		for i = 1, NUM_RAID_GROUPS do
			BENECAST_RAID_LIST['raid'..i] = nil;
		end
		BeneCast_SnapPanels('BeneCast');
	elseif ( event == 'ADDON_LOADED' and arg1 == 'Blizzard Raid UI' ) then
		BeneCast_UpdateRaid();
		BeneCast_SnapPanels('BeneCast');
	elseif ( event == 'ADDON_LOADED' ) then
		-- Place the panels in the correct place
		BeneCast_SnapPanels(arg1);
	elseif ( event == 'RAID_ROSTER_UPDATE' ) then
		BeneCast_UpdateRaid();
	end
	
end

-- *****************************************************************************
-- BeneCastMiniMapButton Game Event Handlers
-- *****************************************************************************

-- Clicks on the BeneCast button do different things
function BeneCastOptionFrameToggle(mouseButton)

	-- If the mouseButton is the left button toggle BeneCastFrame's visibility
	if ( mouseButton == 'LeftButton' ) then
		local frame = getglobal('BeneCastOptionFrame');
		if ( frame:IsVisible() ) then
			frame:Hide();
		else
			frame:Show();
		end
	end
	
end

function BeneCastMiniMap_OnMouseDown()

	if ( ( BeneCastConfig.Unlock ) and ( arg1 == 'LeftButton' ) ) then
		this:StartMoving();
		this.isMoving = true;
	end

end

function BeneCastMiniMap_OnLoad()

	this:RegisterEvent('ADDON_LOADED');

end

function BeneCastMiniMap_OnEvent()

	if ( event == 'ADDON_LOADED' and arg1 == 'BeneCast' ) then
		if ( BeneCastConfig.HideMinimap ) then
			getglobal('BeneCastMinimapButton'):Hide();
		end
	end

end